import os
import logging
import atexit
import signal
from pathlib import Path


# ----------------------------------------------------------------------
def autokill_process():
    """"""
    pid = os.getpid()
    logging.info(f"Running with PID: {pid}")

    ID_FILE = os.path.join(Path.home(), ".openbci.pid")
    logging.info(f"PID file created in: {ID_FILE}")

    if os.path.exists(ID_FILE):
        with open(ID_FILE, 'r') as file:
            oldpid = file.read()
            if oldpid.isdigit():
                oldpid = int(oldpid)
                if oldpid != pid:
                    try:
                        os.kill(oldpid, signal.SIGKILL)
                    except:
                        pass

    with open(ID_FILE, 'w') as file:
        file.write(str(pid))

    # ----------------------------------------------------------------------
    @atexit.register
    def _():
        logging.info(f"Killing PID: {pid}")
        try:
            os.kill(-pid, signal.SIGKILL)
            if os.path.exists(ID_FILE):
                os.remove(ID_FILE)
        except:
            pass




# import psutil
import logging

from openbci.stream.ws.server import create_server
from openbci.utils import autokill_process

autokill_process()
# from openbci.acquisition import Cyton

# threading.stack_size(2**26)
# sys.setrecursionlimit(2**20)

logging.getLogger().setLevel(logging.INFO)
logging.info(f"EEGSTREAM started.")


# minimal configuration
config = {
    'websocket_port': 8845,
    # 'shield_wifi_ip': '192.168.1.113',
    # 'sample_rate': Cyton.SAMPLE_RATE_1KHZ,
    # 'sample_rate': Cyton.SAMPLE_RATE_250HZ,
    # 'sample_rate': Cyton.SAMPLE_RATE_500HZ,
}


# ----------------------------------------------------------------------
def run():
    create_server(config)


if __name__ == '__main__':
    run()




"""
Brython MDCComponent: MDCDrawer
===============================


"""

from browser import html
from .core import MDCTemplate


########################################################################
class __drawerItem__(MDCTemplate):
    """"""
    # NAME = 'draweritem', 'Drawer Item'

    MDC_optionals = {

        'ignore_link': 'RDNT-ignore_link',
        'icon': '<i class="material-icons mdc-list-item__graphic" aria-hidden="true">{icon}</i>',
        'fa_icon': '<i class="mdc-list-item__graphic {fa_style} {fa_icon}"></i>',

    }

    # ----------------------------------------------------------------------

    def __new__(self, name, icon=None, link='#', ignore_link=False, item_theme='', **kwargs):
        """"""
        if icon and icon.startswith('fa'):
            fa_style = icon[:icon.find('-')]
            fa_icon = 'fa' + icon[icon.find('-'):]
            del icon

        self.element = self.render(locals(), kwargs)

        return self.element

    # ----------------------------------------------------------------------

    @classmethod
    def __html__(cls, **context):
        """"""
        code = """
        <a class="mdc-list-item {ignore_link}" href="{link}">
            {icon}
            {fa_icon}
            <span class="mdc-list-item__text">{name}</span>
        </a>
        """
        return cls.render_html(code, context)


########################################################################
class MDCDrawer(MDCTemplate):
    """"""

    CSS_classes = {
        # 'permanent': 'mdc-drawer--permanent',
        # 'fixed':  'mdc-toolbar--fixed',
        # 'waterfall': 'mdc-toolbar--waterfall',
        # 'flexible': 'mdc-toolbar--flexible',
        # 'fixed_lastrow_only': 'mdc-toolbar--fixed-lastrow-only',
    }

    # ----------------------------------------------------------------------
    def __new__(self, header='', mode='modal', theme='primary', item_theme=None, title='', subtitle='', **kwargs):
        """"""

        self.theme = theme
        # self.item_theme = item_theme

        if mode == 'modal':
            self.element = self.render(locals(), kwargs, attach_now=False)
            self.element.style = {'min-height': '100%'}
            self.element.attach = self.attach
        else:
            self.element = self.render(locals(), kwargs)
            self.element.style = {'min-height': '100%'}
            # self.element.attach = self.attach

        self.element.mode = mode

        return self.element

    # ----------------------------------------------------------------------

    @classmethod
    def __html__(cls, **context):
        """"""

        context['theme_'] = {'primary': 'secondary',
                             'secondary': 'primary', }[context['theme']]

        # context['item_theme_'] = {'primary': 'secondary',
                                  # 'secondary': 'primary', }[context['item_theme']]

        cls.NAME = 'drawer', 'MDCDrawer'

        if context.get('mode') == 'permanent':
            code = """
                <aside class="mdc-drawer mdc-drawer--permanent">
                
                  <div class="mdc-drawer__header">
                    <h3 class="mdc-drawer__title">{title}</h3>
                    <h6 class="mdc-drawer__subtitle">{subtitle}</h6>
                  </div>
                
                  <div class="mdc-drawer__content">
                    <nav class="mdc-list">
                
                    </nav>
                  </div>
                </aside>
            """
        elif context.get('mode') == 'dismissible':
            code = """
                <aside class="mdc-drawer mdc-drawer--dismissible">
                
                  <div class="mdc-drawer__header">
                    <h3 class="mdc-drawer__title">{title}</h3>
                    <h6 class="mdc-drawer__subtitle">{subtitle}</h6>
                  </div>
                
                  <div class="mdc-drawer__content">
                    <nav class="mdc-list">
                
                    </nav>
                  </div>
                </aside>
            """

        elif context.get('mode') == 'modal':
            code = """
                <aside class="mdc-drawer mdc-drawer--modal">
                
                  <div class="mdc-drawer__header">
                    <h3 class="mdc-drawer__title">{title}</h3>
                    <h6 class="mdc-drawer__subtitle">{subtitle}</h6>
                  </div>
                
                  <div class="mdc-drawer__content">
                    <nav class="mdc-list">
                
                    </nav>
                  </div>
                </aside>
                
 
            """

        return cls.render_html(code, context)

    # ----------------------------------------------------------------------

    @classmethod
    def __getitem__(self, name):
        """"""
        try:

            if name is 'content':
                return self.element.select('.mdc-list')[0]
            elif name is 'items':
                return self.element.select('.mdc-list .mdc-list-item')
            elif name is 'header':
                return self.element.select('.mdc-drawer__header')[0]
            elif name is 'drawer_drawer':
                return self.element.select('.mdc-drawer__drawer')[0]

            # elif name is 'title':
                # return self.element.select('.mdc-toolbar__title')[0]

        except IndexError:
            return None

    # ----------------------------------------------------------------------

    @classmethod
    def add_item(cls, element, name, icon=None, link='#', **kwargs):
        """"""
        kwargs['item_theme'] = 'mdc-theme--{}'.format(cls.theme)

        item = __drawerItem__(name, icon, link, **kwargs)
        print("IMHERE78")
        #item.bind('click', cls.__activate__(item))
        #item.bind('click', item.mdc.activate)

        cls['content'] <= item
        return item

    # ----------------------------------------------------------------------
    @classmethod
    def add_divider(cls, element, **kwargs):
        """"""
        kwargs['Class'] = 'mdc-list-divider ' + kwargs.get('Class', '')
        divider = html.HR(**kwargs)
        cls['content'] <= divider
        return divider

    # ----------------------------------------------------------------------
    @classmethod
    def add_subheader(cls, element, label, **kwargs):
        """"""
        kwargs['Class'] = 'mdc-list-group__subheader ' + kwargs.get('Class', '')
        subheader = html.H6(label, **kwargs)
        cls['content'] <= subheader
        return subheader

    # ----------------------------------------------------------------------
    # @classmethod
    # def __activate__(self, item):
        # """"""
        #inset = lambda evt:item.mdc.activate()
        # return inset

    # ----------------------------------------------------------------------
    # @classmethod
    # def activate(cls, item, *args, **kwargs):
        # """"""
        # print('EWE')

        # for it in document.select('.mdc-drawer .mdc-list-item'):
            #it.class_name = it.class_name.replace('mdc-list-item--activated', '')

        # item.mdc.add_class(['mdc-list-item--activated'])

    # ----------------------------------------------------------------------

    @classmethod
    def open(cls, element):
        """"""
        cls.mdc.open = True

    # ----------------------------------------------------------------------

    @classmethod
    def close(cls, element):
        """"""
        cls.mdc.open = False

    # ----------------------------------------------------------------------
    @classmethod
    def toggle(cls, element):
        """"""
        # print(cls.mdc.open)
        cls.mdc.open = not cls.mdc.open




"""
Brython MDCComponent: MDCTab
============================


"""


from .core import MDCTemplate

from browser import html, window
# from .MDCButton import MDCButton,  MDCIconToggle


########################################################################
class __tabItem__(MDCTemplate):
    """"""

    NAME = 'tab', 'MDCTab'

    MDC_optionals = {

        # 'outlined': 'mdc-card--outlined',
        # 'full_bleed': 'mdc-card__actions--full-bleed',
        'icon': '<span class="mdc-tab__icon material-icons" aria-hidden="true">{icon}</span>',

        # 'fa_icon': '<i class="mdc-button__icon {fa_style} {fa_icon}"></i>',
        'fa_icon': '<span class="mdc-tab__icon {fa_style} {fa_icon}"></span>',


        'active': 'mdc-tab--active',
        'indicator_active': 'mdc-tab-indicator--active',
        'stacked': 'mdc-tab--stacked',

    }

    # ----------------------------------------------------------------------

    def __new__(self, text=None, icon=None, id='#', stacked=False, active=False, **kwargs):
        """"""

        indicator_active = active

        if icon and icon.startswith('fa'):
            fa_style = icon[:icon.find('-')]
            fa_icon = 'fa' + icon[icon.find('-'):]
            del icon

        self.element = self.render(locals(), kwargs)
        self.element.bind('click', self.bind_click(self.element))

        self.element.indicator = window.mdc.tabIndicator.MDCTabIndicator.attachTo(self.element.select(".mdc-tab-indicator")[0])

        # self.element.bind('click', lambda ev: [button.deactivate() for button in self.element.parent.children])
        # self.element.bind('click', lambda ev: self.element.indicator.activate())

        # self.element.deactivate = self.element.indicator.deactivate

        return self.element

    # ----------------------------------------------------------------------
    @classmethod
    def bind_click(self, element):
        """"""
        def inset(event):
            [button.indicator.deactivate() for button in element.parent.children]
            element.indicator.activate()
            element.clicked()

        return inset

    # ----------------------------------------------------------------------

    @classmethod
    def __html__(cls, **context):
        """"""
        if context['text'] and not (context.get('icon', False) or context.get('fa_icon', False)):
            code = """
            <button class="mdc-tab {active}" role="tab" aria-selected="true">
              <span class="mdc-tab__content">
                <span class="mdc-tab__text-label">{text}</span>
              </span>

              <span class="mdc-tab-indicator {indicator_active}">
                <span class="mdc-tab-indicator__content mdc-tab-indicator__content--underline"></span>
              </span>

              <span class="mdc-tab__ripple"></span>
            </button>
            """

        elif not context['text'] and (context.get('icon', False) or context.get('fa_icon', False)):
            code = """
            <button class="mdc-tab {active}" role="tab" aria-selected="true">
              <span class="mdc-tab__content">
                {icon}
                {fa_icon}
              </span>
              <span class="mdc-tab-indicator {indicator_active}">
                <span class="mdc-tab-indicator__content mdc-tab-indicator__content--underline"></span>
              </span>
              <span class="mdc-tab__ripple"></span>
            </button>
            """

        elif context['text'] and (context.get('icon', False) or context.get('fa_icon', False)):
            code = """
            <button class="mdc-tab {active} {stacked}" role="tab" aria-selected="true">
              <span class="mdc-tab__content">
                {icon}
                {fa_icon}
                <span class="mdc-tab__text-label">{text}</span>
              </span>
              <span class="mdc-tab-indicator {indicator_active}">
                <span class="mdc-tab-indicator__content mdc-tab-indicator__content--underline"></span>
              </span>
              <span class="mdc-tab__ripple"></span>
            </button>
            """

        return cls.render_html(code, context)

    # ----------------------------------------------------------------------

    @classmethod
    def __getitem__(self, name):
        """"""
        if name is 'label':
            return self.element.select('.mdc-tab__text-label')[0]

        # elif name is 'items_location':
            # try:
                # return self.element.select('.mdc-tab-scroller__scroll-content')[0]
            # except:
                # return self.element


########################################################################
class MDCTabBar(MDCTemplate):
    """"""

    NAME = 'tabBar', 'MDCTabBar'

    CSS_classes = {

        # '_16_9': 'mdc-card__media--16-9',
        # 'square': 'mdc-card__media--square',
    }

    MDC_optionals = {

        # 'outlined': 'mdc-card--outlined',
        # 'full_bleed': 'mdc-card__actions--full-bleed',
        # 'icon': '<i class="material-icons mdc-button__icon" aria-hidden="true">{icon}</i>',
        # 'disabled': 'disabled',

    }

    # ----------------------------------------------------------------------

    def __new__(self, *items, **kwargs):
        """"""
        self.element = self.render(locals(), kwargs)
        self.element.panels = html.DIV(Class='panels')
        # self.element.mdc.bind('MDCTabBar:activated', self.__tabchanged__(self.element))

        self.element.stacked = kwargs.get('stacked', False)

        self.element.panel = {}

        for item in items:
            _, panel = self.add_item(self.element, **item)
            self.element.panel[item['id']] = panel

        return self.element

    # ----------------------------------------------------------------------
    @classmethod
    def __tabchanged__(cls, element, id, item):
        """"""
        # print('HHH', element.mdc, id)

        for e in element.select(".mdc-tab--active"):
            e.class_name = e.class_name.replace('mdc-tab--active', '')

        item.class_name += ' mdc-tab--active'

        try:

            for panel in element.panels.select('.panel'):
                panel.style = {'display': None, }

            # index = element.mdc.getFocusedTabIndex()
            element.panel[id].style = {'display': 'block', }

        except:
            pass
        # return inset

    # ----------------------------------------------------------------------

    @classmethod
    def __html__(cls, **context):
        """"""
        code = """
    <div class="mdc-tab-bar" role="tablist">
    <div class="mdc-tab-scroller">
    <div class="mdc-tab-scroller__scroll-area">
      <div class="mdc-tab-scroller__scroll-content">

      </div>
      </div>
      </div>
      </div>
        """

        return cls.render_html(code, context)

    # ----------------------------------------------------------------------

    @classmethod
    def __getitem__(self, name):
        """"""
        if name is 'items':
            return self.element.select('.mdc-tab')

        elif name is 'items_location':
            try:
                return self.element.select('.mdc-tab-scroller__scroll-content')[0]
            except:
                return self.element

        # elif name is 'action_icons':
            # return self.element.select('.mdc-card__action-icons')[0]

    # ----------------------------------------------------------------------

    @classmethod
    def add_item(cls, element, text=None, icon=None, id='#', active=False):
        """"""
        item = __tabItem__(text=text, icon=icon, id=id, active=active, stacked=element.stacked)
        item.clicked = lambda: cls.__tabchanged__(element, id, item)
        # item.bind('click', lambda event: cls.__tabchanged__(element, id, item))
        cls['items_location'] <= item

        if active:
            active = 'active'
            display = 'block'
        else:
            active = ''
            display = None

        panel = html.DIV(Class='panel {}'.format(active), id=id, role='tabpanel', aria_hidden=True)
        panel.style = {'display': display, }

        cls.element.panel[id] = panel
        cls.element.panels <= panel

        return item, panel

    # ----------------------------------------------------------------------
    # @classmethod
    # def title(self, mdc, text):
        # """"""
        # self['title'].text = text


########################################################################
class MDCTabScroller(MDCTabBar):
    """"""

    NAME = 'tabScroller', 'MDCTabScroller'

    # ----------------------------------------------------------------------

    @classmethod
    def __html__(cls, **context):
        """"""
        code = """
            <div class="mdc-tab-scroller">
              <div class="mdc-tab-scroller__scroll-area">
                <div class="mdc-tab-scroller__scroll-content"></div>
              </div>
            </div>
        """

        return cls.render_html(code, context)





"""
===========
HTTP Server
===========

Tornado HTTP server with a WebSocket running in it, it is composed of:

  * **Application**: A collection of request handlers that make up a web application.
  * **HTTPServer**: A non-blocking, single-threaded HTTP server.
"""

from threading import Thread
# from multiprocessing import Process
import logging
import os

from openbci.stream.ws.app import EndPointHandler, HomeHandler
from openbci.stream.ws.base_server import WSHandler_Serial, WSHandler_WiFi

from tornado.httpserver import HTTPServer
from tornado.ioloop import IOLoop
from tornado.web import Application, url

DEBUG = True
STATIC_DIRNAME = "static"


# ----------------------------------------------------------------------
def make_app(config):
    """A collection of request handlers that make up a web application.

    Parameters
    ----------
    config: dict
        Object with configurations from main window.

    Returns
    -------
    Application
        A collection of request handlers that make up a web application.
    """

    settings = {
        'debug': DEBUG,
        'xsrf_cookies': False,

        "static_path": os.path.join(os.path.dirname(__file__), STATIC_DIRNAME),
        "static_url_prefix": "/static/",
    }

    config['download'] = 'http://localhost:{}{}'.format(config['websocket_port'], settings['static_url_prefix'])

    return Application([

        url(r'^/', HomeHandler),
        url(r'^/api', EndPointHandler),

        url(r'^/wifi', WSHandler_WiFi, {'bci_config': config, }),
        url(r'^/serial', WSHandler_Serial, {'bci_config': config, }),

    ], **settings)


# ----------------------------------------------------------------------
def create_server(config):
    """Entry point for invote the Websocket from main window.

    Parameters
    ----------
    config: dict
        Object with configurations from main window.

    Returns
    -------
    HTTPServer
        Server object that can be stoped in the main window.
    """

    port = int(config['websocket_port'])

    logging.info("OpenBCI server running on port {}".format(port))
    application = make_app(config)
    http_server = HTTPServer(application)

    try:
        http_server.listen(port)
    except:
        pass

    if hasattr(IOLoop.instance(), 'asyncio_loop'):

        # Start new IOLoop only if previous is dead or not exist
        if not IOLoop.instance().asyncio_loop.is_running():
            try:
                # Using a thread for keep main window alive
                Thread(target=IOLoop.instance().start, args=()).start()
                # Process(target=IOLoop.instance().start, args=()).start()
            except:
                pass  # TODO: Windows test

    else:  # Sometimes this instance ('asyncio_loop') not exist in windows.
        Thread(target=IOLoop.instance().start, args=()).start()
        # Process(target=IOLoop.instance().start, args=()).start()

    return http_server  # returning server to main window


# ----------------------------------------------------------------------
def stop_websocket_server():
    """Try to stop server.

    `Try` means that the task will be released to OS.
    """

    ioloop = IOLoop.current()
    ioloop.add_callback(ioloop.stop)


if __name__ == '__main__':

    config = {
        'websocket_port': 8845,
        'user_dir': '.',
    }
    create_server(config)




from .events import get_event_loop


class InvalidStateError(Exception):
    pass


class CancelledError(Exception):
    pass


class TimeoutError(Exception):
    pass


class Future:
    """
        A class representing the future result of an async action.
        Implementations should override the :method:`start` method
        which should start the asynchronous operation. The class will
        typically register a handler to be run when the operation finishes.
        This handler then needs to call the base :method:`_finish` method
        providing it with the :parameter:`result` parameter and
        :parameter:`status` (which should either be ``Promise.STATUS_FINISHED``
        in case the operation finished successfully or ``Promise.STATUS_ERROR``
        if an error happened).
    """
    STATUS_STARTED = 0
    STATUS_CANCELED = 1
    STATUS_FINISHED = 2
    STATUS_ERROR = 3

    def __init__(self, loop=None):
        if loop is None:
            loop = get_event_loop()
        self._loop = loop
        self._status = Future.STATUS_STARTED
        self._result = None
        self._exception = None
        self._callbacks = []
        self._loop._register_future(self)

    def _schedule_callbacks(self):
        for cb in self._callbacks:
            self._loop.call_soon(cb, self)

    def cancel(self):
        """
        Cancel the future and schedule callbacks.

        If the future is already done or cancelled, return False. Otherwise, change the future’s state to cancelled, schedule the callbacks and return True."""
        if self._status != Future.STATUS_STARTED:
            return False
        self._status = Future.STATUS_CANCELED
        self._schedule_callbacks()
        return True

    def cancelled(self):
        """Return True if the future was cancelled."""
        return self._status == Future.STATUS_CANCELED

    def done(self):
        """
        Return True if the future is done.

        Done means either that a result / exception are available, or that the future was cancelled.
        """
        return self._status != Future.STATUS_STARTED

    def result(self):
        """
        Return the result this future represents.

        If the future has been cancelled, raises CancelledError. If the future’s result isn’t yet available, raises InvalidStateError. If the future is done and has an exception set, this exception is raised.
        """
        if self._status == Future.STATUS_STARTED:
            raise InvalidStateError()
        if self._status == Future.STATUS_CANCELED:
            raise CancelledError()
        if self._status == Future.STATUS_ERROR:
            raise self._exception
        return self._result

    def exception(self):
        """
        Return the exception that was set on this future.

        The exception (or None if no exception was set) is returned only if the future is done. If the future has been cancelled, raises CancelledError. If the future isn’t done yet, raises InvalidStateError.
        """
        if self._status == Future.STATUS_STARTED:
            raise InvalidStateError()
        if self._status == Future.STATUS_CANCELED:
            raise CancelledError()
        if self._status == Future.STATUS_ERROR:
            return self._exception

    def add_done_callback(self, fn):
        """
        Add a callback to be run when the future becomes done.

        The callback is called with a single argument - the future object. If the future is already done when this is called, the callback is scheduled with call_soon().

        Use functools.partial to pass parameters to the callback. For example, fut.add_done_callback(functools.partial(print, "Future:", flush=True)) will call print("Future:", fut, flush=True).
        """
        if self.done():
            self._loop.call_soon(fn,self)
        else:
            self._callbacks.append(fn)

    def remove_done_callback(self, fn):
        """
        Remove all instances of a callback from the “call when done” list.

        Returns the number of callbacks removed.
        """
        removed = 0
        retain = []
        for cb in self._callbacks:
            if cb == fn:
                removed += 1
            else:
                retain.append(cb)
        self._callbacks = retain
        return removed

    def set_result(self, result):
        """
        Mark the future done and set its result.

        If the future is already done when this method is called, raises InvalidStateError.
        """
        if self._status != Future.STATUS_STARTED:
            raise InvalidStateError()
        self._result = result
        self._status = Future.STATUS_FINISHED
        self._schedule_callbacks()

    def set_exception(self, exception):
        """
        Mark the future done and set an exception.

        If the future is already done when this method is called, raises InvalidStateError.
        """
        if self._status != Future.STATUS_STARTED:
            raise InvalidStateError()
        self._exception = exception
        self._status = Future.STATUS_ERROR
        self._schedule_callbacks()

    def __iter__(self):
        if not self.done():
            yield self
        return self.result()


class GatheredFuture(Future):
    def __init__(self, futures, return_exceptions=False, loop=None):
        super().__init__(loop=None)
        self._futures = futures
        self._ret_exceptions = return_exceptions
        for fut in futures:
            fut.add_done_callback(lambda : self._done(fut))

    def _done(self, fut):
        if self.done():
            return

        if fut.canceled():
            if not self._ret_exceptions:
                self.set_exception(CancelledError())
                return
        else:
            exc = fut.exception()
            if exc is not None and not self._ret_exceptions:
                self.set_exception(exc)
                return
        self._check_finished()

    def cancel(self):
        for fut in self._futures:
            if not fut.done():
                fut.cancel()
        super().cancel()

    def _check_finished(self):
        results = []
        for fut in self._futures:
            if not fut.done():
                return
            elif fut.canceled():
                results.append(CancelledError())
            exc = fut.exception()
            if exc is not None:
                results.append(exc)
            else:
                results.append(fut.result())
        self.set_result(results)


class SleepFuture(Future):
    def __init__(self, seconds, result=None, loop=None):
        super().__init__(loop)
        self._loop.call_later(seconds, self.set_result, result)

    def set_result(self, result):
        if not self.done():
            super().set_result(result)
        else:
            print("Sleep already finished with ex:", self.exception())


def gather(*coros_or_futures, return_exceptions=False, loop=None):
    fut_list = [ensure_future(c, loop=loop) for c in coros_or_futures]
    return GatheredFuture(fut_list, return_exceptions=False)



""" The asyncio package, tracking PEP 3156.
    A module providing the :class:`Future` and :function:`coroutine` decorator
    which can be make async operations look like sync. Typical usecase is
    as follows. Assume we have a method ``query_server`` which asynchronously
    queries a server. So, instead of returning the results, it returns an
    instance of the :class:`Future` class. Normally one would call
    the :method:`then` method of this instance providing it with a call back
    to be called when the promise is "resolved", i.e. when the query has
    finished and results are ready. This, however, typically leads to something
    called the "callback" hell. The :function:`coroutine` decorator can get
    around this by a clever use of the ``yield`` statement. So, instead of
    writing:

    ```
        def process_results(results):
            do_some_stuff(results)

        def send_query():
            promise = query_server()
            promise.then(process_results)
    ```

    one can write it in a more straightforward way:

    ```
        @coroutine
        def process_query():
            results = yield from query_server()
            do_some_stuff(results)
    ```

    eliminating the need to introduce the ``process_results`` callback.
"""

from .coroutines import *
from .events import *
from .futures import *
from .locks import *
from .queues import *

from .http import *
from .objects import *


FIRST_COMPLETED = 0
FIRST_EXCEPTION = 0
ALL_COMPLETED = 0



def wait_for(coro_or_future, timeout, *args, loop=None):
    if loop is None:
        loop = get_event_loop()
    fut = ensure_future(coro_or_future(*args), loop=loop)
    if timeout:
        def timeout_handler():
            if fut.done():
                pass
            else:
                fut.set_exception(TimeoutError())
        loop.call_later(timeout, timeout_handler)
    return fut

def shield(arg, loop=None):
    """Wait for a future, shielding it from cancellation.

    The statement

        res = yield from shield(something())

    is exactly equivalent to the statement

        res = yield from something()

    *except* that if the coroutine containing it is cancelled, the
    task running in something() is not cancelled.  From the POV of
    something(), the cancellation did not happen.  But its caller is
    still cancelled, so the yield-from expression still raises
    CancelledError.  Note: If something() is cancelled by other means
    this will still cancel shield().

    If you want to completely ignore cancellation (not recommended)
    you can combine shield() with a try/except clause, as follows:

        try:
            res = yield from shield(something())
        except CancelledError:
            res = None
    """
    inner = ensure_future(arg)
    if inner.done():
        # Shortcut.
        return inner
    if loop is None:
        loop = get_event_loop()
    outer = Future()

    def _done_callback(inner):
        if outer.cancelled():
            if not inner.cancelled():
                # Mark inner's result as retrieved.
                inner.exception()
            return

        if inner.cancelled():
            outer.cancel()
        else:
            exc = inner.exception()
            if exc is not None:
                outer.set_exception(exc)
            else:
                outer.set_result(inner.result())

    inner.add_done_callback(_done_callback)
    return outer

def sleep(seconds, result=None, loop=None):
    return futures.SleepFuture(seconds, result, loop=loop)



"""
Brython MDCComponent: MDCDialog
===============================


"""

from .core import MDCTemplate
from .MDCButton import MDCButton

########################################################################


class MDCDialog(MDCTemplate):
    """"""

    NAME = 'dialog', 'MDCDialog'

    CSS_classes = {

        # '_16_9': 'mdc-card__media--16-9',
        # 'square': 'mdc-card__media--square',
    }

    MDC_optionals = {

        'scrollable': 'mdc-dialog--scrollable',
        # 'full_bleed': 'mdc-card__actions--full-bleed',
        # 'icon': '<i class="material-icons mdc-button__icon" aria-hidden="true">{icon}</i>',
        # 'disabled': 'disabled',

    }

    # ----------------------------------------------------------------------
    def __new__(self, title='', content='', scrollable=False, **kwargs):
        """"""
        self.element = self.render(locals(), kwargs)
        return self.element

    # ----------------------------------------------------------------------

    @classmethod
    def __html__(cls, **context):
        """"""
        cls.ID = cls.new_id()
        context['id'] = cls.ID

        code = """
        
        <div class="mdc-dialog"
             role="alertdialog"
             aria-modal="true"
             id="{id}"
             aria-labelledby="{id}-title"
             aria-describedby="{id}-content">
          <div class="mdc-dialog__container">
            <div class="mdc-dialog__surface">
              <h2 class="mdc-dialog__title" id="{id}-title">{title}</h2>
              <div class="mdc-dialog__content" id="{id}-content">
                {content}
              </div>
              <footer class="mdc-dialog__actions">
              </footer>
            </div>
          </div>
          <div class="mdc-dialog__scrim"></div>
        </div>        
        """

        return cls.render_html(code, context)

    # ----------------------------------------------------------------------
    @classmethod
    def __getitem__(self, name):
        """"""
        if name is 'actions':
            return self.element.select('.mdc-dialog__actions')[0]

        elif name is 'surface':
            return self.element.select('.mdc-dialog__surface')[0]

        elif name is 'content':
            return self.element.select('.mdc-dialog__content')[0]

        elif name is 'title':
            return self.element.select('.mdc-dialog__title')[0]

        # elif name is 'action_icons':
            # return self.element.select('.mdc-card__action-icons')[0]

    # ----------------------------------------------------------------------

    @classmethod
    def add_footer_button(cls, element, *args, **kwargs):
        """"""

        button = MDCButton(*args, **kwargs)
        button.class_name += ' mdc-dialog__button'

        if kwargs.get('action', False):
            button.attrs['data-mdc-dialog-action'] = kwargs['action']

        if kwargs.get('default', False):
            button.class_name += ' mdc-dialog__button--default'

        # if kwargs.get('cancel', False):
            # button.class_name += ' mdc-dialog__footer__button--cancel'
        # elif kwargs.get('accept', False):
            # button.class_name += ' mdc-dialog__footer__button--accept'

        # if kwargs.get('action', False):
            # button.class_name += ' mdc-dialog__action'

        if kwargs.get('theme', False):
            button.mdc.theme(kwargs['theme'])

        cls['actions'] <= button
        return button

    # ----------------------------------------------------------------------
    # @classmethod
    # def show(cls, *args, **kwargs):
        # """"""
        # dialog = window.mdc.dialog.MDCDialog.new(document.querySelector('#{}'.format(cls.ID)))
        # dialog.show()

    # ----------------------------------------------------------------------
    # @classmethod
    # def close(cls, *args, **kwargs):
        # """"""
        # dialog = window.mdc.dialog.MDCDialog.new(document.querySelector('#{}'.format(cls.ID)))
        # dialog.close()

    # ----------------------------------------------------------------------
    # @classmethod
    # def open(cls, *args, **kwargs):
        # """"""
        # dialog = window.mdc.dialog.MDCDialog.new(document.querySelector('#{}'.format(cls.ID)))
        # return dialog.open



"""
=========
WS Server
=========

A Python client for access to WebSocket streammed data.
"""

import sys
import json
import logging
import numpy as np
from datetime import datetime

from socket import error as socket_error
from abc import ABCMeta, abstractmethod
import time

from tornado.ioloop import IOLoop, PeriodicCallback
from tornado import gen
from tornado.websocket import websocket_connect

from multiprocessing import Process, Manager, Queue
from queue import Queue as ClassicQueue


# MARKERS = Queue(maxsize=1)
# FINISH = Queue(maxsize=1)
MARKERS = Manager().Queue(maxsize=1)
FINISH = Manager().Queue(maxsize=1)


########################################################################
class WSClient(metaclass=ABCMeta):
    """Base client for headwares implementations.

    This class just connect with server and redirect the data flow.
    """

    # ----------------------------------------------------------------------
    def start(self, ip, device_id, port, mode='wifi', pack_size=1000, boardmode='BOARD_MODE_MARKER', sample_rate='SAMPLE_RATE_1KHZ', montage=None, force=False):
        """Try to connect with the WebSocket server.

        On WebSocket mode the acquisition only works with package data.

        Parameters
        ----------
        ip : str
            Is the IP where run WebSocket the server.
        port : str
            The port for the WebSocket server.
        mode: str, optional
            The board mode must be specified explicitly, the options are:
            `serial` and `wifi` (default).
        pack_size : int, optional
            Milliseconds for packs.
        boardmode : str, optional
            The boardmode, one of `BOARD_MODE_DEFAULT`, `BOARD_MODE_DEBUG`,
            `BOARD_MODE_ANALOG`, `BOARD_MODE_DIGITAL`, `BOARD_MODE_MARKER` or
            `BOARD_MODE_GET`.
        """

        self.url = f'ws://{ip}:{port}/{mode}'
        self.ioloop = IOLoop.instance()

        self.connect()
        self.port = port
        self.ip = ip
        self.pack_size = pack_size
        self.boardmode = boardmode
        self.sample_rate_ = sample_rate
        self.device_id = device_id
        self.montage = montage
        self.force = force

        self.ioloop.start()

    # # ----------------------------------------------------------------------
    # @gen.coroutine
    # def _check_markers(self):
        # """"""
        # logging.warning("MMM")
        # while not FINISH.qsize():
        # pass
        # logging.warning("W")
        # if MARKERS.qsize():
        # logging.warning("MARKER")
        # marker, burst = MARKERS.get()
        # yield self.ws.write_message(json.dumps({'command': 'send_marker', 'kwargs': {'marker': marker, 'burst': 4, }, }))
        # time.sleep(0.001)

    # ----------------------------------------------------------------------
    def __getattr__(self, attr):
        """For call remote methods with local syntax.

        The WebSocket server consist in a set of methods that can be called with
        a JSON structure, for example `start_stream`, this method need
        `boardmode` and `milliseconds` arguments`, the WebSocket message must be
        costructed as:

        >>> message = "{'command':'start_stream','kwargs':{'boardmode':'BOARD_MODE_MARKER','milliseconds':1000}}"
        >>> self.ws.write_message(message)

        for prevent this unfamiliar structure, this method add a Python like
        syntax:

        >>> self.start_stream(boardmode='BOARD_MODE_MARKER', milliseconds=1000)

        even if the `start_stream` method is not explicitly defined.


        Parameters
        ----------
        attr : str
            Funtion name

        Returns
        -------
        function
            Artificial function that send a request to the WebSocket.
        """

        def inset(**kwargs):
            self.ws.write_message(json.dumps(
                {'command': attr, 'kwargs': kwargs, }))
        return inset

    # ----------------------------------------------------------------------
    @gen.coroutine
    def connect(self):
        """Try connection with server.

        The streamming will start automatically.

        Returns
        -------
        bool
            `True` if coonected, otherwise `False`.
        """

        logging.info(f"Trying connection to...")
        try:
            self.ws = yield websocket_connect(self.url)
        except Exception as e:
            logging.error(f"Connection error:{self.url}, {e}")
            sys.exit()
        else:
            logging.info(f"Connected to {self.url}")
            self._run()
            # self._check_markers()

    # ----------------------------------------------------------------------
    @gen.coroutine
    def _run(self):
        """Start the data streamming.

        This method will call a `callback` method on each received message.
        """

        if self.force:
            self.set_up(device_id=self.device_id,
                        sample_rate=self.sample_rate_, montage=self.montage)
        else:
            self.register(device_id=self.device_id,
                          sample_rate=self.sample_rate_, montage=self.montage)
        self.start_stream(milliseconds=self.pack_size,
                          boardmode=self.boardmode)

        while not FINISH.qsize():

            if MARKERS.qsize():
                marker, burst = MARKERS.get()
                # logging.warning(f"MARKER: {marker} {burst}")
                yield self.ws.write_message(json.dumps({'command': 'send_marker', 'kwargs': {'marker': marker, 'burst': burst, }, }))

            data = yield self.ws.read_message()

            if not data:
                logging.warning(f'No data in pack')
                logging.warning(f'{data}')
                continue

            data = json.loads(data)

            if not 'data' in data:
                continue

            self.save_data(data)
            self.callback(self, data)

        yield self.ws.write_message(json.dumps({'command': 'stop_stream', 'kwargs': {}, }))
        self.close()

    # ----------------------------------------------------------------------
    def close(cls):
        ioloop = IOLoop.current()
        ioloop.add_callback(ioloop.stop)

    # ----------------------------------------------------------------------
    @abstractmethod
    def callback(self, parent, data):
        """Callback method for each data package.

        Parameters
        ----------
        data : dict
            The packaged deserialized data from headware.
        """
        pass


########################################################################
class _CytonWS(WSClient):
    """Use `WSClient` for redirect the data to a Queue."""

    # ----------------------------------------------------------------------
    def __init__(self, ip, device_id, port=8845, mode='wifi', queue=None, pack_size=1000, boardmode='BOARD_MODE_MARKER', callback=None, sample_rate='SAMPLE_RATE_1KHZ', montage=[], force=False):
        """Example function with types documented in the docstring.

        `PEP 484`_ type annotations are supported. If attribute, parameter, and
        return types are annotated according to `PEP 484`_, they do not need to be
        included in the docstring:

        Parameters
        ----------
        ip : str
            Is the IP where run WebSocket the server.
        port : str
            The port for the WebSocket server.
        mode: str, optional
            The board mode must be specified explicitly, the options are:
            `serial` and `wifi` (default).
        pack_size : int, optional
            Milliseconds for packs.
        boardmode : str, optional
            The boardmode, one of `BOARD_MODE_DEFAULT`, `BOARD_MODE_DEBUG`,
            `BOARD_MODE_ANALOG`, `BOARD_MODE_DIGITAL`, `BOARD_MODE_MARKER` or
            `BOARD_MODE_GET`.
        queue : queue, ooptional
             If a queue is provided the system will use it, otherwise a new one
             will be created.
        """

        if queue:
            self._eeg_pack = queue
        else:
            self._eeg_pack = Manager().Queue()

        self._eeg_buffer = Manager().Queue()

        self.pack_size = pack_size

        if callback:
            self.callback = callback
        else:
            logging.info("No callback function defined!!")

        self._start_collect = lambda pack_size: self.start(
            ip, device_id, port, mode, pack_size, boardmode, sample_rate, montage, force)

    # ----------------------------------------------------------------------
    def start_collect(self, pack_size=None):
        """"""
        if pack_size is None:
            pack_size = self.pack_size
        self._start_collect(pack_size)

    # ----------------------------------------------------------------------
    def start_collect_process(self, pack_size=None):
        """"""
        if pack_size is None:
            pack_size = self.pack_size
        self.collect_process = Process(
            target=self._start_collect, args=(pack_size, ))
        self.collect_process.start()
        return self.collect_process

    # ----------------------------------------------------------------------
    def stop_collect_process(self):
        """"""
        FINISH.put("dummy")
        self.collect_process.terminate()
        self.collect_process.join()
        self.dump_queues()

    # ----------------------------------------------------------------------
    @property
    def eeg_pack(self):
        """property handler for the main queue."""

        return self._eeg_pack

    # ----------------------------------------------------------------------
    @property
    def eeg_buffer(self):
        """"""
        return self._eeg_buffer

    # ----------------------------------------------------------------------
    def dump_queues(self):
        """"""
        _eeg_pack = ClassicQueue()
        _eeg_buffer = ClassicQueue()

        while self._eeg_pack.qsize():
            _eeg_pack.put(self._eeg_pack.get())

        while self._eeg_buffer.qsize():
            _eeg_buffer.put(self._eeg_buffer.get())

        self._eeg_buffer = _eeg_buffer
        self._eeg_pack = _eeg_pack

    # ----------------------------------------------------------------------
    # @gen.coroutine
    def callback(self, parent, data):
        """Callback method for each data package.

        Add the input data to the queue.

        Parameters
        ----------
        data : dict
        The packaged deserialized data from headware.
        """

        self

        # self.save_data(data)

    # ----------------------------------------------------------------------

    def save_data(self, data):
        """"""

        self._eeg_pack.put(np.array(data['data']).T)

        for d in np.array(data['data']).T:
            self._eeg_buffer.put(d)

    # ----------------------------------------------------------------------
    @property
    def sample_rate(self):
        """"""
        if self.eeg_buffer.qsize() > 2:
            eeg = np.array(self.eeg_buffer.queue)
            sr = eeg.shape[0] / (datetime.fromtimestamp(eeg[-1][4]) -
                                 datetime.fromtimestamp(eeg[0][4])).total_seconds()
            del eeg
            return sr

        if self.eeg_pack.qsize() >= 2:
            eeg = np.array(self.eeg_pack.queue)[-1]
            sr = eeg.shape[0] / (datetime.fromtimestamp(eeg[-1][4]) -
                                 datetime.fromtimestamp(eeg[0][4])).total_seconds()
            del eeg
            return sr

        else:
            logging.warning(f"Not enough data for calculate the sample frequency.")


########################################################################
class CytonWS_dummy:

    # ----------------------------------------------------------------------
    def __init__(self, ws):
        # self.marker = None
        self._eeg_pack = Manager().Queue()
        self.ws = ws

    # ----------------------------------------------------------------------
    @property
    def eeg_pack(self):
        """"""
        return self._eeg_pack

    # ----------------------------------------------------------------------
    def close(cls):
        ioloop = IOLoop.current()
        ioloop.add_callback(ioloop.stop)

    # ----------------------------------------------------------------------
    def send_marker(self, marker, burst=4):
        """"""
        if not MARKERS.qsize():
            MARKERS.put((marker, burst))


# ########################################################################
# class CytonWS:
    # """"""
    # # ----------------------------------------------------------------------
    # def __init__(self, ip, device_id, port=8845, mode='wifi', queue=None, pack_size=1000, boardmode='BOARD_MODE_MARKER', callback=None, sample_rate='SAMPLE_RATE_1KHZ', montage=[], force=False):
        # """Constructor"""

        # self.openbci = _CytonWS(ip, device_id, port=port, mode=mode, queue=queue, pack_size=pack_size, boardmode=boardmode, callback=callback, sample_rate=sample_rate, montage=montage, force=force)

    # # ----------------------------------------------------------------------
    # def collect(self, t1, wait=True):
        # """"""
        # self.start_collect()
        # if wait:
            # self.wait_for_data()
        # time.sleep(t1)
        # self.openbci.stop_collect_process()

    # # ----------------------------------------------------------------------
    # def wait_for_data(self):
        # """"""
        # logging.info("Waiting for data...")
        # while not self.openbci.eeg_pack.qsize():
            # time.sleep(0.1)
        # logging.info("Stream started.")

    # # ----------------------------------------------------------------------
    # def start_collect(self):
        # """"""
        # self.openbci.start_collect_process()

    # # ----------------------------------------------------------------------
    # def stop_collect(self):
        # """"""
        # self.openbci.stop_collect_process()

    # # ----------------------------------------------------------------------
    # def send_marker(self, marker, burst=4):
        # """"""
        # if not MARKERS.qsize():
            # MARKERS.put((marker, burst))

    # # ----------------------------------------------------------------------
    # def __getattr__(self, attr):
        # """"""

        # if hasattr(self.openbci, attr):
            # return getattr(self.openbci, attr)


########################################################################
class CytonWS(_CytonWS):
    """

    Main methonds like `start_collect` run inside a Process.
    """

    # ----------------------------------------------------------------------
    def collect(self, t1, wait=True):
        """"""
        self.start_collect(self.pack_size)
        if wait:
            self.wait_for_data()
        time.sleep(t1)
        self.stop_collect_process()

    # ----------------------------------------------------------------------
    def wait_for_data(self):
        """"""
        logging.info("Waiting for data...")
        while not self.eeg_pack.qsize():
            time.sleep(0.1)
        logging.info("Stream started.")

    # ----------------------------------------------------------------------
    def start_collect(self, pack_size=None):
        """"""
        if pack_size:
            pack_size = self.pack_size
        return self.start_collect_process(pack_size)

    # ----------------------------------------------------------------------
    def stop_collect(self):
        """"""
        self.stop_collect_process()

    # ----------------------------------------------------------------------
    def send_marker(self, marker, burst=4):
        """"""
        if not MARKERS.qsize():
            MARKERS.put((marker, burst))


# ----------------------------------------------------------------------
def CytonWS_decorator(ip, device_id, port=8845, mode='wifi', queue=None, pack_size=1000, boardmode='BOARD_MODE_MARKER', sample_rate='SAMPLE_RATE_1KHZ', montage=[], force=False):
    """"""
    def wrapper(fn):
        def my_callback(ws, data):
            fn(ws, data)

        ws = _CytonWS(ip, device_id, port, mode, pack_size=pack_size, boardmode=boardmode,
                      callback=my_callback, sample_rate=sample_rate, montage=montage, force=force)
        _ws = CytonWS_dummy(ws)
        ws._eeg_pack = _ws.eeg_pack
        ws.start_collect(pack_size)

    return wrapper



"""Event loop."""

from browser import ajax
from browser import timer

import asyncio
import time


class Handle:
    def __init__(self):
        self._cancelled = False

    def cancel(self):
        self._cancelled = True

    def cancelled(self):
        return self._cancelled


class TimerHandle(Handle):
    """Object returned by timed callback registration methods."""

    def __init__(self, timeout, when):
        self._cancelled = False
        self._timeout = timeout
        self._when = when


    def __hash__(self):
        return hash(self._when)

    def __lt__(self, other):
        return self._when < other._when

    def __le__(self, other):
        if self._when < other._when:
            return True
        return self.__eq__(other)

    def __gt__(self, other):
        return self._when > other._when

    def __ge__(self, other):
        if self._when > other._when:
            return True
        return self.__eq__(other)

    def __eq__(self, other):
        if isinstance(other, TimerHandle):
            return self._timeout == other._timeout
        return NotImplemented

    def __ne__(self, other):
        equal = self.__eq__(other)
        return NotImplemented if equal is NotImplemented else not equal

    def cancel(self):
        if not self._cancelled:
            timer.clear_timeout(self._timeout)
            self._cancelled = True


class PausableTimerHandle(TimerHandle):
    def __init__(self, callback, when, schedule_immediately=True):
        super().__init__(None, when)
        self._finished = False
        def _cb():
            self._finished = True
            callback()
        self._cb = _cb
        self._cb.__name__ = callback.__name__
        if schedule_immediately:
            self.resume()

    def resume(self):
        if not self._finished and not self._timeout:
            tm = time.time()
            if tm >= self._when:
                delay = 0
            else:
                delay = (self._when-tm)*1000
            self._timeout = timer.set_timeout(self._cb, delay)

    def pause(self):
        if not self._finished and self._timeout:
            timer.clear_timeout(self._timeout)

    def cancel(self):
        if not self._cancelled and not self._finished:
            if self._timeout:
                timer.clear_timeout(self._timeout)
            self._cancelled = True
            self._finished = True

    def finished(self):
        return self._finished


class BrowserEventLoop:
    def __init__(self):
        self._debug = False
        self._pending =[]
        self._running=True
        self._scheduled = []
        self._cleanup_timer = timer.set_interval(self._clean_schedule, 1000)

    def _clean_schedule(self):
        updated = []
        for handle in self._scheduled:
            if not handle.finished():
                updated.append(handle)
        self._scheduled = updated

    def _register_future(self, fut):
        self._pending.append(fut)
        fut.add_done_callback(self._future_done)

    def _future_done(self, fut):
        self._pending.remove(fut)

    def pending_futures(self):
        return self._pending

    def send_exception_to_pending_futures(self, ex):
        for fut in self._pending:
            try:
                if not fut.done():
                    fut.set_exception(ex)
            except BaseException as exc:
                console.log("Error sending exception", ex, "to", fut, ":", exc)

    def close(self):
        self.stop()
        if self._debug:
            self._clean_schedule()
            if len(self._scheduled) > 0:
                print("Warning, pending callbacks", self._scheduled)
            if len(self._pending) > 0:
                print("Warning, pending futures", self._pending)

    def stop(self, *args):
        if self._running:
            for handle in self._scheduled:
                handle.pause()
            self._clean_schedule()
            timer.clear_interval(self._cleanup_timer)
            self._cleanup_timer = None
            self._running = False

    def run_forever(self):
        if not self._running:
            for handle in self._scheduled:
                handle.resume()
            self._clean_schedule()
            self._cleanup_timer = timer.set_interval(self._clean_schedule, 1000)
            self._running = True

    def run_until_complete(self, coro_or_fut):
        if not self._running:
            self.run_forever()
        fut = asyncio.ensure_future(coro_or_fut, loop=self)
        fut.add_done_callback(self.stop)
        return fut

    def is_running(self):
        return self._running

    def call_soon(self, callback, *args):
        return self.call_at(self.time(), callback, *args)

    def call_later(self, delay, callback, *args):
        return self.call_at(self.time()+delay, callback, *args)

    def call_at(self, when, callback, *args):
        def _callback():
            callback(*args)
        _callback.__name__ = callback.__name__
        handle = PausableTimerHandle(_callback, when, schedule_immediately = self._running)
        self._scheduled.append(handle)
        return handle

    def time(self):
        return time.time()

    def get_debug(self):
        return self._debug

    def set_debug(self, enabled):
        self._debug = enabled


class AbstractEventLoopPolicy:
    def __init__(self):
        loop = self.new_event_loop()
        self.set_event_loop(loop)

    def get_event_loop(self):
        return self._loop

    def set_event_loop(self, loop):
        self._loop = loop

    def new_event_loop(self):
        return BrowserEventLoop()


_default_policy = AbstractEventLoopPolicy()
_current_policy = _default_policy


def get_event_loop_policy():
    return _current_policy


def set_event_loop_policy(policy):
    global current_policy
    _current_policy = policy or _default_policy


def get_event_loop():
    return _current_policy.get_event_loop()


def set_event_loop(loop):
    _current_policy.set_event_loop(loop)


def new_event_loop():
    return _current_policy.new_event_loop()







"""
Brython MDCComponent: MDCLayoutGrid
===================================


"""

from .core import MDCTemplate
#from .MDCButton import MDCButton,  MDCIconToggle



########################################################################
class __cell__(MDCTemplate):
    """"""

    NAME = 'cell', 'MDCCell'


    MDC_optionals = {

        'desktop': 'mdc-layout-grid__cell--span-{desktop}-desktop',
        'tablet': 'mdc-layout-grid__cell--span-{tablet}-tablet',
        'mobile': 'mdc-layout-grid__cell--span-{mobile}-mobile',
        'C': 'mdc-layout-grid__cell--span-{C}',

        'index': 'mdc-layout-grid__cell--order-{index}',
        'position': 'mdc-layout-grid__cell--align-{position}',

        'fixed_column_width': 'mdc-layout-grid--fixed-column-width',


    }



    #----------------------------------------------------------------------
    @classmethod
    def __html__(cls, **context):
        """"""
        code = """
            <div class="mdc-layout-grid__cell {C} {desktop} {tablet} {mobile} {index} {position} {fixed_column_width}"></div>
        """
        return cls.render_html(code, context)




########################################################################
class __grid__(MDCTemplate):
    """"""
    NAME = 'grid', 'MDCGrid'


    MDC_optionals = {

    'position': 'mdc-layout-grid--align-{position}',

    }

    #----------------------------------------------------------------------
    @classmethod
    def __html__(cls, **context):
        """"""
        code = """
            <div class="mdc-layout-grid__inner {position}">
            </div>
        """
        return cls.render_html(code, context)


    #----------------------------------------------------------------------
    @classmethod
    def cell(cls, element, **kwargs):
        """"""
        cell = __cell__(**kwargs)
        cls.element <= cell
        return cell




########################################################################
class MDCLayoutGrid(MDCTemplate):
    """"""

    NAME = 'layoutgrid', 'MDCLayoutGrid'

    CSS_classes = {

        #'_16_9': 'mdc-card__media--16-9',
        #'square': 'mdc-card__media--square',
    }

    MDC_optionals = {

        #'outlined': 'mdc-card--outlined',
        #'full_bleed': 'mdc-card__actions--full-bleed',
        #'icon': '<i class="material-icons mdc-button__icon" aria-hidden="true">{icon}</i>',
        #'disabled': 'disabled',

    }

    #----------------------------------------------------------------------
    def __new__(self, **kwargs):
        """"""
        self.element = self.render(locals(), kwargs)
        return self.element



    #----------------------------------------------------------------------
    @classmethod
    def __html__(cls, **context):
        """"""
        code = """
            <div class="mdc-layout-grid">

            </div>
        """


        return cls.render_html(code, context)




    #----------------------------------------------------------------------
    @classmethod
    def __getitem__(self, name):
        """"""
        #if name is 'actions':
            #return self.element.select('.mdc-card__actions')[0]

        #elif name is 'action_buttons':
            #return self.element.select('.mdc-card__action-buttons')[0]

        #elif name is 'action_icons':
            #return self.element.select('.mdc-card__action-icons')[0]


    #----------------------------------------------------------------------
    @classmethod
    def grid(cls, element, *args, **kwargs):
        """"""
        grid = __grid__(*args, **kwargs)
        cls.element <= grid
        return grid



    ##----------------------------------------------------------------------
    #@classmethod
    #def title(self, mdc, text):
        #""""""
        #self['title'].text = text





# import sys
# import struct
# from functools import reduce
# from operator import add

# VREF = 4.5
# GAIN = 24
# FACTOR_SCALE = VREF / (24 * ((2 ** 23) - 1))

# input_ = sys.argv[1]

# input_file = open(input_, 'r')
# output_file = open(input_.replace('raw', 'bin'), 'wb')

# line = input_file.readline()
# sample = 0

# while line:
    # data_eeg = reduce(add, [list(struct.pack("I", int(float(n) // FACTOR_SCALE))[:3]) for n in line.split(',')[0:9]])

    # pre = ['A0', hex(sample % 255)]
    # end = ['00', '00', '00', '00', '00', '00', 'C0']
    # data = [int(v, 16) for v in pre] + data_eeg + [int(v, 16) for v in end]

    # output_file.write(bytearray(data))
    # line = input_file.readline()
    # sample += 1


# output_file.close()
# input_file.close()



"""
=====
Cyton
=====

The OpenBCI Cyton PCBs were designed with Design Spark, a free PCB capture
program.


Cyton Board Specs:

* Power with 3-6V DC Battery ONLY
* PIC32MX250F128B Micrcontroller with chipKIT UDB32-MX2-DIP bootloader
* ADS1299 Analog Front End
* LIS3DH 3 axis Accelerometer
* RFduino BLE radio
* Micro SD card slot
* Voltage Regulation (3V3, +2.5V, -2.5V)
* Board Dimensions 2.41 x 2.41 (octogon has 1 edges) [inches]
* Mount holes are 1/16 ID, 0.8 x 2.166 on center [inches]


Data Format
===========

Binary Format
-------------

+-------------+----------------------------------------------------------------+
| **Byte No** | **Description**                                                |
+-------------+----------------------------------------------------------------+
| 1           | Start byte, always `0xA0`                                      |
+-------------+----------------------------------------------------------------+
| 2           | Sample Number                                                  |
+-------------+----------------------------------------------------------------+
| 3-26        | EEG Data, values are 24-bit signed, MSB first                  |
+-------------+----------------------------------------------------------------+
| 27-32       | Aux Data                                                       |
+-------------+----------------------------------------------------------------+
| 33          | Footer, `0xCX` where `X` is 0-F in hex                         |
+-------------+----------------------------------------------------------------+



EEG Data for 8 channels
-----------------------

24-Bit Signed.

+-------------+----------------------------------------------------------------+
| **Byte No** | **Description**                                                |
+-------------+----------------------------------------------------------------+
| 3-5         | Data value for EEG channel 1                                   |
+-------------+----------------------------------------------------------------+
| 6-8         | Data value for EEG channel 2                                   |
+-------------+----------------------------------------------------------------+
| 9-11        | Data value for EEG channel 3                                   |
+-------------+----------------------------------------------------------------+
| 12-14       | Data value for EEG channel 4                                   |
+-------------+----------------------------------------------------------------+
| 15-17       | Data value for EEG channel 5                                   |
+-------------+----------------------------------------------------------------+
| 18-20       | Data value for EEG channel 6                                   |
+-------------+----------------------------------------------------------------+
| 21-23       | Data value for EEG channel 7                                   |
+-------------+----------------------------------------------------------------+
| 24-26       | Data value for EEG channel 8                                   |
+-------------+----------------------------------------------------------------+


EEG Data for 16 channels
------------------------

24-Bit Signed.

+----------------------------+--------------------------+--------------------------+
| **Received**               | **Upsampled board data** | **Upsampled daisy data** |
+--------------+-------------+--------------------------+--------------------------+
| sample(3)    |             | avg(sample(1),sample(3)) | sample(2)                |
+--------------+-------------+--------------------------+--------------------------+
|              | sample(4)   | sample(3)                | avg(sample(2),sample(4)) |
+--------------+-------------+--------------------------+--------------------------+
| sample(5)    |             | avg(sample(3),sample(5)) | sample(4)                |
+--------------+-------------+--------------------------+--------------------------+
|              | sample(6)   | sample(5)                | avg(sample(4),sample(6)) |
+--------------+-------------+--------------------------+--------------------------+
| sample(7)    |             | avg(sample(5),sample(7)) | sample(7)                |
+--------------+-------------+--------------------------+--------------------------+
|              | sample(8)   | sample(7)                | avg(sample(6),sample(8)) |
+--------------+-------------+--------------------------+--------------------------+


Aux Data
--------

16-Bit Signed.

+--------------------+-------------+-------------+-------------+-------------+-------------+-------------+-------------------------------+
| **Stop Byte (33)** | **Byte 27** | **Byte 28** | **Byte 29** | **Byte 30** | **Byte 31** | **Byte 32** |          **Name**             |
+--------------------+-------------+-------------+-------------+-------------+-------------+-------------+-------------------------------+
| 0xc0               | AX1         | AX2         | AY1         | AY2         | AZ1         | AZ2         | Standard with accel           |
+--------------------+-------------+-------------+-------------+-------------+-------------+-------------+-------------------------------+
| 0xC1               | UDF         | UDF         | UDF         | UDF         | UDF         | UDF         | Standard with raw aux         |
+--------------------+-------------+-------------+-------------+-------------+-------------+-------------+-------------------------------+
| 0xC2               | UDF         | UDF         | UDF         | UDF         | UDF         | UDF         | User defined                  |
+--------------------+-------------+-------------+-------------+-------------+-------------+-------------+-------------------------------+
| 0xC3               | *AC*        | *AV*        | T3          | T2          | T1          | T0          | Time stamped set with accel   |
+--------------------+-------------+-------------+-------------+-------------+-------------+-------------+-------------------------------+
| 0xC4               | *AC*        | *AV*        | T3          | T2          | T1          | T0          | Time stamped with accel       |
+--------------------+-------------+-------------+-------------+-------------+-------------+-------------+-------------------------------+
| 0xC5               | UDF         | UDF         | T3          | T2          | T1          | T0          | Time stamped set with raw aux |
+--------------------+-------------+-------------+-------------+-------------+-------------+-------------+-------------------------------+
| 0xC6               | UDF         | UDF         | T3          | T2          | T1          | T0          | Time stamped with raw aux     |
+--------------------+-------------+-------------+-------------+-------------+-------------+-------------+-------------------------------+


Aux Data
--------

16-Bit Signed.

+-------------+-------------+
| **Byte 27** | **Byte 28** |
+-------------+-------------+
|  X          | AX1         |
+-------------+-------------+
|  x          | AX0         |
+-------------+-------------+
|  Y          | AY1         |
+-------------+-------------+
|  y          | AY0         |
+-------------+-------------+
|  Z          | AZ1         |
+-------------+-------------+
|  z          | AZ0         |
+-------------+-------------+
"""
import logging
import sys
import serial
import os
import time

import asyncore
import atexit
import socket
import requests

from threading import Thread
from openbci.acquisition.cyton_bin_parser import CytonBinParser

DEFAULT_LOCAL_IP = "192.168.1.1"


########################################################################
class Cyton(CytonBinParser):
    """
    The Cyton data format and SDK define all interactions and capabilities of
    the board, the full instructions can be found in the official package.

      * https://docs.openbci.com/Hardware/03-Cyton_Data_Format
      * https://docs.openbci.com/OpenBCI%20Software/04-OpenBCI_Cyton_SDK

    """

    VREF = 4.5

    BIN_HEADER = 0xa0

    START_STREAM = b'b'
    STOP_STREAM = b's'

    DEACTIVATE_CHANEL = b'12345678qwertyui'
    ACTIVATE_CHANEL = b'!@#$%^&*QWERTYUI'
    CHANNEL_SETTING = b'12345678QWERTYUI'

    TEST_GND = b'0'      # Connect to internal GND (VDD - VSS)
    TEST_1X_SLOW = b'-'  # Connect to test signal 1xAmplitude, slow pulse
    TEST_1X_FAST = b'='  # Connect to test signal 1xAmplitude, fast pulse
    TEST_DC = b'p'       # Connect to DC signal
    TEST_2X_SLOW = b'['  # Connect to test signal 2xAmplitude, slow pulse
    TEST_2X_FAST = b']'  # Connect to test signal 2xAmplitude, fast pulse

    POWER_DOWN_ON = b'0'
    POWER_DOWN_OFF = b'1'

    GAIN_1 = b'0'
    GAIN_2 = b'1'
    GAIN_4 = b'2'
    GAIN_6 = b'3'
    GAIN_8 = b'4'
    GAIN_12 = b'5'
    GAIN_24 = b'6'

    ADSINPUT_NORMAL = b'0'
    ADSINPUT_SHORTED = b'1'
    ADSINPUT_BIAS_MEAS = b'2'
    ADSINPUT_MVDD = b'3'
    ADSINPUT_TEMP = b'4'
    ADSINPUT_TESTSIG = b'5'
    ADSINPUT_BIAS_DRP = b'6'
    ADSINPUT_BIAS_DRN = b'7'

    BIAS_REMOVE = b'0'
    BIAS_INCLUDE = b'1'

    SRB2_DISCONNECT = b'0'
    SRB2_CONNECT = b'1'

    SRB1_DISCONNECT = b'0'
    SRB1_CONNECT = b'1'

    DEFAULT_CHANNELS_SETTINGS = b'd'
    DEFAULT_CHANNELS_SETTINGS_REPORT = b'D'

    TEST_SIGNAL_NOT_APPLIED = b'0'
    TEST_SIGNAL_APPLIED = b'1'

    SD_DATA_LOGGING_5MIN = b'A'
    SD_DATA_LOGGING_15MIN = b'S'
    SD_DATA_LOGGING_30MIN = b'F'
    SD_DATA_LOGGING_1HR = b'G'
    SD_DATA_LOGGING_2HR = b'H'
    SD_DATA_LOGGING_4HR = b'J'
    SD_DATA_LOGGING_12HR = b'K'
    SD_DATA_LOGGING_24HR = b'L'
    SD_DATA_LOGGING_14S = b'a'
    SD_DATA_LOGGING_STOP = b'j'

    QUERY_REGISTER = b'?'
    SOFT_RESET = b'v'

    USE_8CH_ONLY = b'c'
    USE_16CH_ONLY = b'C'

    TIME_STAMP_ON = b'<'
    TIME_STAMP_OFF = b'>'

    SAMPLE_RATE_16KHZ = b'~0'
    SAMPLE_RATE_8KHZ = b'~1'
    SAMPLE_RATE_4KHZ = b'~2'
    SAMPLE_RATE_2KHZ = b'~3'
    SAMPLE_RATE_1KHZ = b'~4'
    SAMPLE_RATE_500HZ = b'~5'
    SAMPLE_RATE_250HZ = b'~6'
    SAMPLE_RATE_GET = b'~~'

    SAMPLE_RATE_VALUE = {
        b'~0': 16e3,
        b'~1': 8e3,
        b'~2': 4e3,
        b'~3': 2e3,
        b'~4': 1e3,
        b'~5': 500,
        b'~6': 250,
    }

    BOARD_MODE_DEFAULT = b'/0'  # Sends accelerometer data in aux bytes
    BOARD_MODE_DEBUG = b'/1'    # Sends serial output over the external serial
                                # port which is helpful for debugging.
    BOARD_MODE_ANALOG = b'/2'   # Reads from analog pins A5(D11), A6(D12) and
                                # if no wifi shield is present, then A7(D13)
                                # as well.
    BOARD_MODE_DIGITAL = b'/3'  # Reads from analog pins D11, D12 and D17.
                                # If no wifi present then also D13 and D18.
    BOARD_MODE_MARKER = b'/4'   # Turns accel off and injects markers
                                # into the stream by sending `X where X is any
                                # char to add to the first AUX byte.
    BOARD_MODE_GET = b'//'      # Get current board mode

    WIFI_SHIELD_ATTACH = b'{'
    WIFI_SHIELD_DEATTACH = b'}'
    WIFI_SHIELD_STATUS = b':'
    WIFI_SHIELD_RESET = b';'

    VERSION = b'V'

    # ----------------------------------------------------------------------
    @property
    def gain(self):
        """Vector with the gains for each channel."""

        # TODO
        # A method for change the ganancy of each channel must be writed here
        return [24, 24, 24, 24, 24, 24, 24, 24]

    # ----------------------------------------------------------------------
    @property
    def scale_factor_eeg(self):
        """Vector with the correct factors for scale eeg data samples."""

        return [self.VREF / (gain * ((2 ** 23) - 1)) for gain in self.gain]

    # ----------------------------------------------------------------------
    @property
    def bin_header(self):
        """Byte with start header."""

        return self.BIN_HEADER

    # # ----------------------------------------------------------------------
    # @property
    # def sample_rate(self):
        # """Byte with start header."""

        # return self.sample_rate_

    # ----------------------------------------------------------------------
    def start(self):
        """Start stream."""

        return self.command(self.START_STREAM)

    # ----------------------------------------------------------------------
    def stop(self):
        """Stop stream."""

        return self.command(self.STOP_STREAM)

    # ----------------------------------------------------------------------
    def deactivate_channel(self, channels):
        """Deactivate the channels specified."""

        [self.command(self.DEACTIVATE_CHANEL[ch]) for ch in channels]

    # ----------------------------------------------------------------------
    def activate_channel(self, channels):
        """Activate the channels specified."""

        chain = ''.join([chr(self.ACTIVATE_CHANEL[ch]) for ch in channels]).encode()
        self.command(chain)
        # [self.command(self.ACTIVATE_CHANEL[ch]) for ch in channels]

    # ----------------------------------------------------------------------
    def command(self, c):
        """Send a command to device.

        Before send the commmand the input buffer is cleared, and after that
        waits 300 ms for a response.

        Parameters
        ----------
        c : bytes
            Command to send.

        Returns
        -------
        str
            If the command generate a response this will be sended back.
        """

        self.reset_input_buffer()
        self.write(c)
        time.sleep(0.3)
        response = self.read(2**11)
        logging.info(f'From device {response}')

        if hasattr(response, 'encode'):
            response = response.encode()
        elif isinstance(response, (list)):
            response = ''.join([chr(r) for r in response]).encode()

        return response

    # ----------------------------------------------------------------------
    def channel_setting(self, channels,
                        power_down=POWER_DOWN_ON,
                        gain=GAIN_24,
                        input_type=ADSINPUT_NORMAL,
                        bias=BIAS_INCLUDE,
                        srb2=SRB2_CONNECT,
                        srb1=SRB1_DISCONNECT):
        """Channel Settings commands.

        Parameters
        ----------
        channels : list
            List of channels that will share the configuration specified.
        power_down : bytes, optional
            POWER_DOWN_ON (default), POWER_DOWN_OFF.
        gain: bytes, optional
            GAIN_24 (default), GAIN_12, GAIN_8, GAIN_6, GAIN_4, GAIN_2, GAIN_1.
        input_type : bytes, optional
            Select the ADC channel input source: ADSINPUT_NORMAL (default),
            ADSINPUT_SHORTED, ADSINPUT_BIAS_MEAS, ADSINPUT_MVDD, ADSINPUT_TEMP,
            ADSINPUT_TESTSIG, ADSINPUT_BIAS_DRP, ADSINPUT_BIAS_DRN,
        bias : bytes, optional
            Select to include the channel input in BIAS generation:
            BIAS_INCLUDE (default), BIAS_REMOVE.
        srb2 : bytes, optional
            Select to connect this channel’s P input to the SRB2 pin. This
            closes a switch between P input and SRB2 for the given channel,
            and allows the P input also remain connected to the ADC:
            SRB2_CONNECT (default), SRB2_DISCONNECT.
        srb1 : bytes, optional
            Select to connect all channels’ N inputs to SRB1. This effects all
            pins, and disconnects all N inputs from the ADC:
            SRB1_DISCONNECT (default), SRB1_CONNECT.

        Returns
        -------

        On success:

          * If streaming, no confirmation of success. Note: WiFi Shields will always get a response, even if streaming.
          * If not streaming, returns Success: Channel set for 3$$$, where 3 is the channel that was requested to be set.

        On failure:

          * If not streaming, NOTE: WiFi shield always sends the following responses without $$$
              * Not enough characters received, Failure: too few chars$$$ (example user sends x102000X)
              * 9th character is not the upper case `X`, Failure: 9th char not X$$$ (example user sends x1020000V)
              * Too many characters or some other issue, Failure: Err: too many chars$$$
          * If not all commands are not received within 1 second, Timeout processing multi byte message - please send all commands at once as of v2$$$

        """

        start = b'x'
        end = b'X'

        self.reset_input_buffer()

        for channel in channels:
            ch = self.CHANNEL_SETTING[channel]
            chain = b''.join([start, ch, power_down, gain, input_type, bias,
                              srb2, srb1, end])
            self.write(chain)

        time.sleep(0.3)
        return self.read(2**11)

    # ----------------------------------------------------------------------
    def leadoff_impedance(self, channels, pchan=TEST_SIGNAL_NOT_APPLIED,
                          nchan=TEST_SIGNAL_NOT_APPLIED):
        """LeadOff Impedance Commands

        Parameters
        ----------
        channels : list
            List of channels that will share the configuration specified.
        pchan : bytes, optional
            TEST_SIGNAL_NOT_APPLIED (default), TEST_SIGNAL_NOT_APPLIED.
        nchan : bytes, optional
            TEST_SIGNAL_NOT_APPLIED (default), TEST_SIGNAL_NOT_APPLIED.

        Returns
        -------

        On success:

          * If streaming, no confirmation of success. Note: WiFi Shields will always get a response, even if streaming.
          * If not streaming, returns Success: Lead off set for 4$$$, where 4 is the channel that was requested to be set.

        On failure:

          * If not streaming, NOTE: WiFi shield always sends the following responses without $$$
              * Not enough characters received, Failure: too few chars$$$ (example user sends z102000Z)
              * 5th character is not the upper case ‘Z’, Failure: 5th char not Z$$$ (example user sends z1020000X)
              * Too many characters or some other issue, Failure: Err: too many chars$$$
          * If not all commands are not received within 1 second, Timeout processing multi byte message - please send all commands at once as of v2$$$

        """

        start = b'z'
        end = b'Z'

        self.reset_input_buffer()

        for channel in channels:
            ch = chr(self.CHANNEL_SETTING[channel]).encode()
            chain = b''.join([start, ch, pchan, nchan, end])
            self.write(chain)

        time.sleep(0.3)
        return self.read(2**11)

    # ----------------------------------------------------------------------
    def get_board_mode(self):
        """The board mode determine how to deserialize the `AUX` bytes.

        Returns
        -------
        str, None
            The board mode as string or None if no response.

        """
        current_mode = self.command(self.BOARD_MODE_GET)
        if not current_mode:
            return None
        for mode in [b'analog', b'digital', b'debug', b'default', b'marker']:
            if mode in current_mode:
                return mode

    # ----------------------------------------------------------------------
    def send_marker(self, marker, burst=20):
        """Send marker to device.

        The marker sended will be added to the `AUX` bytes in the next data
        input.

        The OpenBCI markers does not work as well as expected, so this module
        implement an strategy for make it works. A burst markers are sended but
        just one are readed, this add a limitation: No more that one marker each
        300 ms are permitted.

        Parameters
        ----------
        marker : str, bytes, int
            A single value required.
        burst : int, optional
            How many times the marker will be send.

        """

        def thread_marker(marker):
            if isinstance(marker, int):
                marker = chr(marker)
            elif isinstance(marker, bytes):
                marker = marker.decode()

            for i in range(burst):
                self.write(f'`{marker}'.encode())
                time.sleep(0.007)

        self.thread_marker = Thread(target=thread_marker, args=(marker, ))
        # self.thread_marker = Process(target=thread_marker, args=(marker, ))
        self.thread_marker.start()

    # ----------------------------------------------------------------------
    def daisy(self):
        """Check if a Daisy module is attached.

        This command will activate the Daisy module is this is available.

        Returns
        -------
        bool
            Daisy module activated.

        """
        # self.command(self.SOFT_RESET)  # to update the status information
        response = self.command(self.USE_16CH_ONLY)
        # self.activate_channel(range(16))

        if not response:
            logging.warning(f"Channels no setted correctly")
            return None

        daisy = not (('no daisy to attach' in response.decode()) or ('8' in response.decode()))

        # if self.montage:
            # channels = self.montage.keys()
        if not self.montage and daisy:
            # channels = range(16)
            self.set_montage(range(16))
        elif not self.montage and not daisy:
            # channels = range(8)
            self.set_montage(range(8))

        channels = self.montage.keys()
        self.leadoff_impedance(channels, self.TEST_SIGNAL_NOT_APPLIED, self.TEST_SIGNAL_APPLIED)
        self.activate_channel(channels)

        logging.info(f"Ussing channels: {channels}")

        return daisy

    # ----------------------------------------------------------------------
    def set_montage(self, montage=None):
        """Define the information with que electrodes names.

        If a list is passed the format will be supposed with index as channels,
        the index can be set explicitly with a dictionary instead of a list.

        Parameters
        ----------
        montage : list, dict, None, optional
            Decription of channels used.
        """

        if isinstance(montage, (list, tuple, range)):
            self.montage = {i: ch for i, ch in enumerate(montage)}
        else:
            self.montage = montage


########################################################################
class FakeDevice:
    """For use without a real device connected."""

    # ----------------------------------------------------------------------
    def __init__(self, file):
        """Load a file or open one from path.

        Parameters
        ----------
        file : file object, str
            The file or file path that contain binary data for debug.
        """

        if isinstance(file, str):
            self.file = open(file, 'rb')
        else:
            self.file = file

    # ----------------------------------------------------------------------
    def read(self, size):
        """Fake read function.

        Parameters
        ----------
        size : int
            Size of input buffer.

        Returns
        -------
        bytes
            Data readed.
        """

        time.sleep(0.01)
        return [int.from_bytes(self.file.read(1), "little") for _ in range(size)]

    # ----------------------------------------------------------------------
    def reset_input_buffer(self):
        """Clear input buffer, discarding all that is in the buffer."""
        pass

    # ----------------------------------------------------------------------
    def write(self, data):
        """Output the given data over the serial port."""
        logging.info(f'Serial write: {data.decode()}')

    # ----------------------------------------------------------------------
    def close(self):
        """Close the serial communication."""
        pass


########################################################################
class CytonRFDuino(Cyton):
    """
    RFduino is the default communication mode for Cyton 32 bit, this set a
    serial comunication through a USB dongle with a sample frequency of `250`
    Hz, for 8 or 16 channels.
    """

    # ----------------------------------------------------------------------
    def __init__(self, port=None, montage=None, sample_rate='SAMPLE_RATE_250HZ'):
        """RFduino mode connection.

        Parameters
        ----------
        port : str, optional
            Specific serial port for connection.
        montage: dictl, list, optional
            Decription of channels used.
        timeout: float, optional.
            Read timeout for serial connection.
        write_timeout: float, optional.
            Write timeout for serial connection.
        sample_rate: bytes, optional.
            The command for set the sample rate.
        """

        if not sample_rate in self.SAMPLE_RATE_VALUE.keys():
            sample_rate = getattr(self, sample_rate)

        self.set_montage(montage)

        if port is None:
            port = self._get_free_port()

        if port is None:
            logging.error("No device was auto detected.")
            sys.exit()

        try:
            self.device_ = serial.Serial(port, 115200, timeout=0.5,
                                         write_timeout=0.1,
                                         parity=serial.PARITY_NONE,
                                         stopbits=serial.STOPBITS_ONE)
            self.command(sample_rate)
            super().__init__()
        except Exception as e:
            logging.error(f"Impossible to connect with {port}")
            logging.error(e)
            sys.exit()

    # ----------------------------------------------------------------------
    @property
    def device(self):
        """Device manager."""

        return self.device_

    # ----------------------------------------------------------------------
    def _get_free_port(self):
        """Look for first available port with OpenBCI device.

        Returns
        -------
        str
            String with the port name or None if no ports were founded.
        """

        if os.name == 'nt':
            prefix = 'COM{}',
        elif os.name == 'posix':
            prefix = '/dev/ttyACM{}', '/dev/ttyUSB{}',

        for pref in prefix:
            for i in range(20):
                port = pref.format(i)
                try:
                    d = serial.Serial(port, timeout=0.2)
                    if d.write(self.START_STREAM):
                        d.close()
                        return port
                except:
                    continue

    # ----------------------------------------------------------------------
    def read(self, size):
        """Read size bytes from the serial port.

        Parameters
        ----------
        size : int
            Size of input buffer.

        Returns
        -------
        bytes
            Data readed.
        """

        return self.device.read(size)

    # ----------------------------------------------------------------------
    def write(self, data):
        """Output the given data over the serial port."""

        return self.device.write(data)

    # ----------------------------------------------------------------------
    def reset_input_buffer(self):
        """Clear input buffer, discarding all that is in the buffer."""

        self.device.reset_input_buffer()

    # ----------------------------------------------------------------------
    def reset_output_buffer(self):
        """Clear output buffer, aborting the current output and discarding all that is in the buffer. """
        self.device.reset_output_buffer()

    # ----------------------------------------------------------------------
    def close(self):
        """Close the serial communication."""

        self.device.close()


########################################################################
class CytonWiFi(Cyton):
    """
    This module implement a TCP connection using the WiFi module, this module
    was designed for works with se same syntax that `CytonRFDuino`.
    """

    STREAMING = False

    # ----------------------------------------------------------------------
    def __init__(self, ip_address, montage=None, sample_rate='SAMPLE_RATE_250HZ'):
        """WiFi mode connection.

        Parameters
        ----------
        ip_address: str.
            IP address with for WiFi module.
        montage: dictl, list, optional
            Decription of channels used.
        sample_rate: bytes, optional.
            The command for set the sample rate.

        """

        if not sample_rate in self.SAMPLE_RATE_VALUE.keys():
            sample_rate = getattr(self, sample_rate)
        # self.sample_rate_ = self.SAMPLE_RATE_VALUE[sample_rate]

        self.ip_address = ip_address
        self.stop_tcp_client()

        self.readed = None

        self.local_ip_address = self.get_local_ip_address()

        self.set_montage(montage)

        super().__init__()
        self.create_tcp_server()
        time.sleep(5)  # wait for the server
        self.start_loop()

        # self.start_tcp()

        atexit.register(self.close)
        self.command(sample_rate)

    # ----------------------------------------------------------------------
    def create_tcp_server(self):
        """Create TCP server.

        This server will handle the streaming EEG data.
        """
        self.local_wifi_server = WiFiShieldTCPServer(self.local_ip_address, self._data_serial)
        self.local_wifi_server_port = self.local_wifi_server.socket.getsockname()[1]
        logging.info(f"Opened socket on {self.local_ip_address}:{self.local_wifi_server_port}")

    # ----------------------------------------------------------------------
    def start_tcp_client(self):
        """Connect the board to the TCP server.

        Send configuration of the previous server created to the board, so they
        can connected to.
        """

        if self.ip_address is None:
            raise ValueError('self.ip_address cannot be None')

        logging.info(f"Init WiFi connection with IP: {self.ip_address}")

        self.requests_session = requests.Session()

        # requests.get(f"http://{self.ip_address}/yt")
        response = requests.get(f"http://{self.ip_address}/board")

        if response.status_code == 200:
            board_info = response.json()
            if not board_info['board_connected']:
                raise RuntimeError("No board connected to WiFi Shield.")

        res_tcp_post = requests.post(f"http://{self.ip_address}/tcp",
                                     json={
                                        'ip': self.local_ip_address,
                                        'port': self.local_wifi_server_port,
                                        'output': 'raw',
                                        'delimiter': False,
                                        'latency': 10,
                                         })
        if res_tcp_post.status_code == 200:
            tcp_status = res_tcp_post.json()
            if tcp_status['connected']:
                logging.info("WiFi Shield to Python TCP Socket Established")
            else:
                raise RuntimeWarning("WiFi Shield is not able to connect to local server.")

    # ----------------------------------------------------------------------
    def stop_tcp_client(self):
        """Stop TCP client."""
        requests.delete(f"http://{self.ip_address}/tcp")

    # ----------------------------------------------------------------------
    def get_local_ip_address(self):
        """Connect to internet for get the local IP."""

        try:
            local_ip_address = socket.gethostbyname(socket.gethostname())
            return local_ip_address
        except Exception as e:
            logging.warning(f'{e}\nAssuming {DEFAULT_LOCAL_IP} as local ip address.')
            return DEFAULT_LOCAL_IP

    # ----------------------------------------------------------------------
    def write(self, data):
        """Send command to board through HTTP protocole."""

        if hasattr(data, 'decode'):
            data = data.decode()
        elif isinstance(data, int):
            data = chr(data)

        try:
            logging.info(f"Sending command: '{data}'")
            response = requests.post(f"http://{self.ip_address}/command", json={'command': data})
        except Exception as e:
            logging.info(f"Error on sending command '{data}': {e}")
            return

        if response.status_code == 200:
            self.readed = response.text
        elif response.status_code == 502:
            logging.info(f"No confirmation from board, does not mean fail.")
        else:
            logging.warning(f"Error code: {response.status_code} {response.text}")
            self.readed = None

    # ----------------------------------------------------------------------
    def read(self, size=None):
        """Read the response for some command.

        Not all command return response.
        """

        time.sleep(0.2)  # very important dealy for wait a response.
        return self.readed

    # ----------------------------------------------------------------------
    def close(self):
        """Stop TCP server."""

        if self.STREAMING:
            logging.warning(f"Closing.")
            self.STREAMING = False
            self.stop_collect()
            # self.stop_loop()
            asyncore.close_all()

    # ----------------------------------------------------------------------
    def start_loop(self):
        """Start the TCP server. """
        self.th_loop = Thread(target=asyncore.loop, args=(), )
        # self.th_loop = Process(target=asyncore.loop, args=())
        self.th_loop.start()

    # ----------------------------------------------------------------------
    def start_collect(self, milliseconds=1000, *args, **kwargs):
        """Start a data collection asynchronously.

        Send a command to the board for start the streaming through TCP.

        Parameters
        ----------
        milliseconds: int, optional
            The duration of data for packing.
        """

        self.reset_buffers()
        self.reset_input_buffer()
        self._start_streaming()
        # self._daisy = self.daisy()
        super().start_collect(*args, **kwargs)
        self.start_tcp_client()
        if milliseconds:
            self.pack_data(milliseconds=milliseconds)

    # ----------------------------------------------------------------------
    def stop_collect(self, *args, **kwargs):
        """Stop collecting data."""
        self.READING = False
        self._stop_streaming()

    # ----------------------------------------------------------------------
    def _start_streaming(self):
        """Start streaming."""

        if not self.STREAMING:
            response = requests.get(f"http://{self.ip_address}/stream/start")
            if response.status_code == 200:
                self.STREAMING = True
            else:
                logging.warning(f"Unable to start streaming.\nCheck API for status code {response.status_code} on /stream/start")

    # ----------------------------------------------------------------------
    def _stop_streaming(self):
        """Stop streaming."""

        if self.STREAMING:
            # self.stop_loop()
            response = requests.get(f"http://{self.ip_address}/stream/stop")
            if response.status_code == 200:
                self.STREAMING = False
            else:
                logging.warning(f"Unable to stop streaming.\nCheck API for status code {response.status_code} on /stream/start")

    # ----------------------------------------------------------------------
    def collect_data(self, size=2**8):
        """Load binary data and put in a queue.

        For optimizations issues the data must be read in packages but write one
        to one in a queue, this method must be executed on a thread.

        Parameters
        ----------
        size : int, optional
            The buffer length for read.

        """
        # Flush input buffer, discarding all its contents.
        self.reset_input_buffer()

    # ----------------------------------------------------------------------
    def start(self):
        """This method does not action on WiFi mode."""
        pass

    # ----------------------------------------------------------------------
    def stop(self):
        """This method does not action on WiFi mode."""
        pass

    # ----------------------------------------------------------------------
    def reset_input_buffer(self):
        """Clear input buffer, discarding all that is in the buffer."""
        pass

    # ----------------------------------------------------------------------
    def reset_output_buffer(self):
        """Clear output buffer, aborting the current output and discarding all that is in the buffer. """
        pass


########################################################################
class WiFiShieldTCPServer(asyncore.dispatcher):
    """
    Simple TCP server.
    """

    SIZE = 32
    BUFFER = 10

    # ----------------------------------------------------------------------
    def __init__(self, host, data_queue):
        """Example function with types documented in the docstring.

        Parameters
        ----------
        host : str
            Local IP.
        data_queue : Queue
            Queue object for write stream data.

        """
        asyncore.dispatcher.__init__(self)
        self.create_socket(socket.AF_INET, socket.SOCK_STREAM)

        self.set_reuse_addr()
        self.bind((host, 0))
        self.listen(1)
        self.data = data_queue

    # ----------------------------------------------------------------------
    def handle_accept(self):
        """Redirect the client connection."""
        pair = self.accept()
        if pair is not None:
            sock, addr = pair
            logging.info(f'Incoming connection from {addr}')
            self.handler = asyncore.dispatcher_with_send(sock)
            self.handler.handle_read = self._handle_read
            self.handler.handle_error = self._handle_error

    # ----------------------------------------------------------------------
    def _handle_read(self):
        """Write the input streaming into the Queue object."""
        self.data.extend(self.handler.recv(33 * 90))

    # ----------------------------------------------------------------------
    def _handle_error(self):
        """"""
        # I'm feeling dirty for this "solution"



"""
Brython MDCComponent: MDCGridList
=================================


"""


from .core import MDCTemplate
# from browser import html
#from .MDCButton import MDCButton,  MDCIconToggle


#########################################################################
# class __Tile__(MDCTemplate):

    # ----------------------------------------------------------------------
    # @classmethod
    # def __html__(cls, **context):
        # """"""

        # code = """
            # <li class="mdc-grid-tile">
            # </li>
        # """


########################################################################
class __gridTile__(MDCTemplate):

    NAME = 'gridtile', 'MDCGridTile'

    MDC_optionals = {

        # 'outlined': 'mdc-card--outlined',
        # 'full_bleed': 'mdc-card__actions--full-bleed',
        'icon': '<i class="mdc-grid-tile__icon material-icons">{icon}</i>',
        # 'disabled': 'disabled',
        'img': '<img class="mdc-grid-tile__primary-content" src="{img}"/>',
        'div': '<div class="mdc-grid-tile__primary-content" style="background-image: url({div});"></div>',
        'caption': '<span class="mdc-grid-tile__support-text">{caption}</span>',


    }

    # ----------------------------------------------------------------------
    def __new__(self, title, caption=None, img=None, div=None, theme='primary', **kwargs):
        """"""

        self.theme = theme

        self.element = self.render(locals(), kwargs)
        return self.element

    # ----------------------------------------------------------------------

    @classmethod
    def __html__(cls, **context):
        """"""

        context['theme_'] = {'primary': 'secondary',
                             'secondary': 'primary', }[context['theme']]

        code = """
            <li class="mdc-grid-tile">
              <div class="mdc-grid-tile__primary">
                {img}
                {div}
              </div>
              <span class="mdc-grid-tile__secondary mdc-theme--{theme_}-bg">
                {icon}
                <span class="mdc-grid-tile__title">{title}</span>
                {caption}
              </span>
            </li>
        """

        return cls.render_html(code, context)

    # ----------------------------------------------------------------------
    # @classmethod
    # def new_tile(cls, mdc, *args, **kwargs):
        # """"""
        #tile = __Tile__()
        #cls.element <= tile
        # return tile


########################################################################
class MDCGridList(MDCTemplate):
    """"""

    NAME = 'gridList', 'MDCGridList'

    CSS_classes = {

        # '_16_9': 'mdc-card__media--16-9',
        # 'square': 'mdc-card__media--square',
    }

    MDC_optionals = {

        # 'outlined': 'mdc-card--outlined',
        # 'full_bleed': 'mdc-card__actions--full-bleed',
        'icon_start': 'mdc-grid-list--with-icon-align-start',
        'icon_end': 'mdc-grid-list--with-icon-align-end',
        'aspect_ratio': 'mdc-grid-list--tile-aspect-{aspect_ratio}',
        'caption': 'mdc-grid-list--twoline-caption',
        # 'disabled': 'disabled',

    }

    # ----------------------------------------------------------------------
    def __new__(self, aspect_ratio='16x9', icon_start=False, icon_end=False, **kwargs):
        """"""
        self.element = self.render(locals(), kwargs)
        return self.element

    # ----------------------------------------------------------------------

    @classmethod
    def __html__(cls, **context):
        """"""

        code = """
            <div class="mdc-grid-list {icon_start} {icon_end} {aspect_ratio} {caption}">
              <ul class="mdc-grid-list__tiles">

              </ul>
            </div>
        """

        return cls.render_html(code, context)

    # ----------------------------------------------------------------------

    @classmethod
    def __getitem__(self, name):
        """"""
        # if name is 'actions':
            # return self.element.select('.mdc-card__actions')[0]

        # elif name is 'action_buttons':
            # return self.element.select('.mdc-card__action-buttons')[0]

        # elif name is 'action_icons':
            # return self.element.select('.mdc-card__action-icons')[0]

    # ----------------------------------------------------------------------
    @classmethod
    def grid_tile(cls, element, *args, **kwargs):
        """"""
        grid_tile = __gridTile__(*args, **kwargs)
        cls.element <= grid_tile
        return grid_tile

    # ----------------------------------------------------------------------
    # @classmethod
    # def new_grid_tile(cls, mdc, *args, **kwargs):
        # """"""
        #grid_tile = html.LI(Class='mdc-grid-tile')
        # return grid_tile

    # ----------------------------------------------------------------------
    # @classmethod
    # def title(self, mdc, text):
        # """"""
        #self['title'].text = text





"""
Brython MDCFramework: Radiant
=============================

"""

import random
from browser import ajax, window, websocket
# from tools import Tools
import json


########################################################################
class AndroidMain:
    """AndroidMain

    Connectt with app
    """

    # ----------------------------------------------------------------------
    def __init__(self, url=None, csrftoken=None):
        """Constructor"""

        self.url_ = url

        if csrftoken:
            self.csrftoken = csrftoken

        elif hasattr(window, 'csrftoken'):
            self.csrftoken = window.csrftoken
        else:
            self.csrftoken = "NO_CSRFTOKEN_PROVIDED"

        if url is None:
            # try:
                # self.url_ = '/'
                # self.test()
            # except:
            self.url_ = '/system_python'
            self.test()

    # ----------------------------------------------------------------------
    def __getattr__(self, attr):
        """"""
        if attr.endswith('_async'):
            attr = attr.replace('_async', '')
            f = lambda *args, **kwargs: self.__request_async__(attr, *args, **kwargs)
        else:
            f = lambda *args, **kwargs: self.__request__(attr, *args, **kwargs)

        f.__name__ = attr
        return f

    # ----------------------------------------------------------------------
    def __request__(self, attr, *args, **kwargs):
        """"""
        req = ajax.ajax()
        req.open('POST', self.url_, False)
        req.set_header('content-type', 'application/x-www-form-urlencoded')
        req.send({'name': attr, 'args': args, 'kwargs': kwargs, 'csrfmiddlewaretoken': self.csrftoken})
        return json.loads(req.text)['__RDNT__']

    # ----------------------------------------------------------------------
    def __request_async__(self, attr, *args, **kwargs):
        """"""
        def __ajax__(fn):
            req = ajax.ajax()
            req.bind('complete', fn)
            req.open('POST', self.url_, True)
            req.set_header('content-type', 'application/x-www-form-urlencoded')
            req.send({'name': attr, 'args': args, 'kwargs': kwargs, 'csrfmiddlewaretoken': self.csrftoken})

        return __ajax__


########################################################################
class Exporter(AndroidMain):
    """Exporter

    Cnnectt with app
    """

    # ----------------------------------------------------------------------
    def __init__(self, url='/system_brython_export', csrftoken=None):
        """Constructor"""
        super().__init__(url, csrftoken)


########################################################################
class LocalInterpreter(AndroidMain):
    """"""


########################################################################
class WebSocket:
    """WebSocket

    COnnectt with app

    """
    # ----------------------------------------------------------------------
    def __init__(cls, ip):
        """"""
        if not websocket.supported:
            print("WebSocket is not supported by your browser")
            return
        # open a web socket
        # cls.ws = websocket.WebSocket("wss://192.168.1.20:8888")
        cls.ws = websocket.WebSocket(ip)
        # bind functions to web socket events
        cls.ws.bind('open', cls.on_open)
        cls.ws.bind('error', cls.on_error)
        cls.ws.bind('message', cls.on_message)
        cls.ws.bind('close', cls.on_close)

        print("Connectting to {}".format(ip))

        port_ip = ip.replace('wss://', '').replace('ws://', '').replace('/wifi', '')

        # print(port_ip)

        # cls.ip = port_ip[:port_ip.find(":")]
        # cls.port =  port_ip[port_ip.find(":")+1:]
        cls.ip = port_ip
        cls.protocol = 'wss' if 'wss' in ip else 'ws'

    # ----------------------------------------------------------------------
    def on_open(cls, evt):
        """"""
        print("Websocket engage.")

    # #----------------------------------------------------------------------
    # def on_connected(self, fn):
        # """"""
        # fn()

    # ----------------------------------------------------------------------
    def on_error(cls, evt):
        """"""
        print("Error.")

    # ----------------------------------------------------------------------
    def on_message(cls, evt):
        """"""
        print("Message received: {}".format(evt.data))

    # ----------------------------------------------------------------------
    def on_close(cls, evt):
        """"""
        print("Websocket closed.")

    # ----------------------------------------------------------------------
    def send(cls, data):
        """"""
        if not data:
            return

        if isinstance(data, (str, bytes)):
            cls.ws.send(data)

        else:
            data = json.dumps(data)
            cls.ws.send(data)

    # ----------------------------------------------------------------------
    def close_connection(cls, *args, **kwargs):
        """"""
        cls.ws.close()


########################################################################
class API:
    """"""

    # ----------------------------------------------------------------------
    def __init__(self, url, auth=None, token=None, csrftoken=None, append_slash=True):
        """Constructor"""
        self.url_ = url
        self.url_ = self.url_.strip('/')

        self.token = token

        if csrftoken:
            self.csrftoken = csrftoken

        elif hasattr(window, 'csrftoken'):
            self.csrftoken = window.csrftoken
        else:
            self.csrftoken = None

        self.append_slash = append_slash

    # ----------------------------------------------------------------------
    def head(self, endpoint):
        """"""
        return self.__request__('HEAD', f'{self.url_}/{endpoint}')

    # ----------------------------------------------------------------------

    def get(self, endpoint, pk=None, data=None):
        """"""
        if pk:
            return self.__request__('GET', f'{self.url_}/{endpoint}/{pk}', data)
        else:
            return self.__request__('GET', f'{self.url_}/{endpoint}', data)

    # ----------------------------------------------------------------------
    def delete(self, endpoint, pk, data=None):
        """"""
        return self.__request__('DELETE', f'{self.url_}/{endpoint}/{pk}', data)

    # ----------------------------------------------------------------------
    def post(self, endpoint, data=None):
        """"""
        return self.__request__('POST', f'{self.url_}/{endpoint}', data)

    # ----------------------------------------------------------------------
    def patch(self, endpoint, pk, data=None):
        """"""
        return self.__request__('PATCH', f'{self.url_}/{endpoint}/{pk}', data)

    # ----------------------------------------------------------------------
    def options(self, endpoint, data=None):
        """"""
        return self.__request__('OPTIONS', f'{self.url_}/{endpoint}', data)

    # ----------------------------------------------------------------------
    def __request__(self, method, url, data=None):
        """"""
        req = ajax.ajax()

        if self.append_slash:
            url = f'{url.strip("/")}/'

        req.open(method, url, False)

        req.setRequestHeader('content-type', 'application/json')

        if self.token:
            req.set_header('Authorization', f"JWT {self.token}")

        if self.csrftoken:
            req.setRequestHeader('X-CSRFToken', self.csrftoken)

        if data:
            data = json.dumps(data)
            req.send(data)
        else:
            req.send()

        data = json.loads(req.text)
        return data


# ----------------------------------------------------------------------
def get(url, method="GET", data=None, csrftoken=None):
    """"""
    if csrftoken:
        csrftoken = csrftoken

    elif hasattr(window, 'csrftoken'):
        csrftoken = window.csrftoken
    else:
        csrftoken = None

    req = ajax.ajax()
    req.open(method, url, False)
    req.set_header('content-type', 'application/x-www-form-urlencoded')
    # req.send({'csrfmiddlewaretoken': csrftoken})

    if data:
        data = json.dumps(data)
        req.send(data)
    else:
        req.send()

    return json.loads(req.text)


# ----------------------------------------------------------------------
def gen_id(length=8, charset='abcdefghijklmnopqrstuvwxyz123456789'):
    """"""
    return ''.join([random.choice(charset) for n in range(length)])




"""
Brython MDCComponent: MDCChips
==============================


"""

from .core import MDCTemplate


########################################################################
class MDCChip(MDCTemplate):
    """"""

    NAME = 'chips', 'MDCChip'

    CSS_classes = {
        'selected': 'mdc-chip--selected',
        # 'waterfall': 'mdc-toolbar--waterfall',
        # 'flexible': 'mdc-toolbar--flexible',
        # 'fixed_lastrow_only': 'mdc-toolbar--fixed-lastrow-only',
    }

    # ----------------------------------------------------------------------
    def __new__(self, text, leading=False, trailing=False, selected=False, **kwargs):
        """"""
        self.element = self.render(locals(), kwargs)
        return self.element

    # ----------------------------------------------------------------------
    @classmethod
    def __html__(cls, **context):
        """"""

        if context.get('leading', False):
            code = '''
                <div class="mdc-chip {selected}">
                  <i class="material-icons mdc-chip__icon mdc-chip__icon--leading">{leading}</i>
                  <div class="mdc-chip__text">{text}</div>
                </div>
            '''

        elif context.get('trailing', False):
            code = '''
                <div class="mdc-chip {selected}">
                  <div class="mdc-chip__text">{text}</div>
                  <i class="material-icons mdc-chip__icon mdc-chip__icon--trailing" tabindex="0" role="button">{trailing}</i>
                </div>
            '''

        else:
            code = '''
                <div class="mdc-chip {selected}" tabindex="0">
                  <div class="mdc-chip__text">{text}</div>
                </div>
                  '''

        return cls.render_html(code, context)


########################################################################
class MDCChipSet(MDCTemplate):
    """"""

    NAME = 'chips', 'MDCChipSet'

    CSS_classes = {

        'choice': 'mdc-chip-set--choice',
        'filter': 'mdc-chip-set--filter',
        'input': 'mdc-chip-set--input',

    }

    # ----------------------------------------------------------------------
    def __new__(self, **kwargs):
        """"""
        self.element = self.render(locals(), kwargs)
        return self.element

    # ----------------------------------------------------------------------
    @classmethod
    def __html__(cls, **context):
        """"""
        code = """
            <div class="mdc-chip-set {CSS_classes}">
            </div>
        """

        return cls.render_html(code, context)

    # ----------------------------------------------------------------------

    @classmethod
    def add_chip(cls, element, text, leading=None, trailing=None, selected=False):
        """"""

        # if selected:
            # selected = 'mdc-chip--selected'
        # else:
            # selected = ''

        # if leading:
            # code = '''
                # <div class="mdc-chip {selected}">
                  # <i class="material-icons mdc-chip__icon mdc-chip__icon--leading">{leading}</i>
                  # <div class="mdc-chip__text">{text}</div>
                # </div>
            # '''.format(text=text, leading=leading, selected=selected)
        # elif trailing:
            # code = '''
                # <div class="mdc-chip {selected}">
                  # <div class="mdc-chip__text">{text}</div>
                  # <i class="material-icons mdc-chip__icon mdc-chip__icon--trailing">{trailing}</i>
                # </div>
            # '''.format(text=text, trailing=trailing, selected=selected)

        # else:
            # code = '''
                # <div class="mdc-chip {selected}" tabindex="0">
                  # <div class="mdc-chip__text">{text}</div>
                # </div>
                  # '''.format(text=text, selected=selected)

        # chip = cls.render_str(code)
        chip = MDCChip(text, leading, trailing, selected)
        cls.element <= chip

        return chip

    # ----------------------------------------------------------------------

    @classmethod
    def __getitem__(self, name):
        """"""
        if name is 'selected':
            chips = self.element.select('.mdc-chip--selected')
            return [c.id for c in chips]
        # elif name is 'items':
            # return self.element.select('.mdc-list .mdc-list-item')

        # elif name is 'title':
            # return self.element.select('.mdc-toolbar__title')[0]





import os

files = os.listdir()
otfs = filter(lambda s: s.endswith('.otf'), files)

font_face = """
@font-face {{
  font-family: '{family}';
  src: url('/static/fonts/worksans/{name}') format('truetype');
}}
"""

names = []
with open('worksans.css', 'w') as file:
    for i, name in enumerate(otfs):

        family = name.replace('.otf', '')

        file.write(font_face.format(**locals()))
        print(f"{i+1}: {name}")
        names.append(f'{family}')


print('-' * 70)
print(', '.join(names) + ';')



"""
Brython MDCComponent: MDCCard
=============================


"""

from .core import MDCTemplate
from .MDCButton import MDCButton, MDCIconToggle

########################################################################


class MDCCard(MDCTemplate):
    """"""

    NAME = 'card', 'MDCCcard'

    MDC_optionals = {

        'outlined': 'mdc-card--outlined',
        'square': 'mdc-card__media--square',
        '_16_9': 'mdc-card__media--16-9',
        'full_bleed': 'mdc-card__actions--full-bleed',
        "primary_action": "mdc-card__primary-action",

    }

    # ----------------------------------------------------------------------
    def __new__(self, **kwargs):
        """"""
        self.element = self.render(locals(), kwargs)
        return self.element

    # ----------------------------------------------------------------------
    @classmethod
    def __html__(cls, **context):
        """"""
        code = """
        <div class="mdc-card {outlined}">
          <div class="mdc-card__media {square} {_16_9}">
            <div class="mdc-card__media-content"></div>
          </div>

          <div class="{primary_action} radiant-content"></div>

          <div class="mdc-card__actions {full_bleed}">

            <div class="mdc-card__action-buttons">
            </div>

            <div class="mdc-card__action-icons">
            </div>

          </div>
        </div>
        """

        return cls.render_html(code, context)

    # ----------------------------------------------------------------------
    @classmethod
    def __getitem__(self, name):
        """"""
        if name is 'actions':
            return self.element.select('.mdc-card__actions')[0]

        elif name is 'content':
            return self.element.select('.radiant-content')[0]

        elif name is 'media':
            return self.element.select('.mdc-card__media')[0]

        elif name is 'media_content':
            return self.element.select('.mdc-card__media-content')[0]

        elif name is 'action_buttons':
            return self.element.select('.mdc-card__action-buttons')[0]

        elif name is 'action_icons':
            return self.element.select('.mdc-card__action-icons')[0]

    # ----------------------------------------------------------------------
    @classmethod
    def add_action_button(cls, element, *args, **kwargs):
        """"""
        button = MDCButton(*args, **kwargs)
        button.class_name += ' mdc-card__action mdc-card__action--button'
        cls['action_buttons'] <= button

        return button

    # ----------------------------------------------------------------------
    @classmethod
    def add_action_icontoggle(cls, element, *args, **kwargs):
        """"""
        button = MDCIconToggle(*args, **kwargs)
        button.class_name += ' mdc-card__action mdc-card__action--icon'
        cls['action_icons'] <= button

        return button

    # ----------------------------------------------------------------------
    @classmethod
    def add_action_icon(cls, element, icon, *args, **kwargs):
        """"""
        button = MDCButton(icon=icon, *args, **kwargs)
        button.class_name += ' mdc-card__action mdc-card__action--icon'
        cls['action_icons'] <= button

        return button



"""
    Asynchronous access to local/remote files.

    Usage:

        from browser import document as doc, html
        import asyncio.fs as afs

        # Create an input element & append it to the page
        i = html.INPUT(type='file')
        doc <= i

        # Wait for the user to select a file ...

        # Once the user has selected the file
        # we can open it using the `open` method
        ff = afs.open(i.files[0])

        # The method returns an `asyncio.Future`
        # Once the file is read, we can retrieve the result with
        f = ff.result()

        # The result is an instance of `asyncio.fs.BrowserFile`
        # which inherits from `io.StringIO`
        # so you can access it with `read`, `readlines`,
        # and all the other methods of `io.StringIO`
        print(f.read())

        # It also has an additional method `save`
        # which downloads the file into the Downloads
        # directory or, depending on user settings,
        # opens a file dialog allowing him to choose where
        # to save the file
        f.save()


    Note: The `asyncio.fs.BrowserFile` object keeps the file contents in memory.

    Note: The `asyncio.fs.BrowserFile` object supports the `write` method. If you
    modify the file using this method, the `save` method will save
    the modified contents.

    Note: The `asyncio.fs.open` method also accepts urls. In that case it will
    download the url and return the contents as an `asyncio.fs.BrowserFile` object.
    However, the `save` method will still save the file locally, **not** remotely.


    The module also contains a convenience method

    `asyncio.fs.open_local`

    Which opens a file dialog which the user can use to pick a file. The method
    returns an `asyncio.Future` which resolves into an `asyncio.fs.BrowserFile`
    with the file contents when the file is read. The method works by
    appending an input to the document body, registering a change handler for it,
    programmatically clicking it, and then immediately deleting it from the document.
"""
import io
import base64

from browser import window, document as doc, html

from .futures import Future
from .http import HTTPRequest
from .coroutines import run_async


class BrowserFile(io.StringIO):
    def __init__(self, name, content='', mime_type='text/plain;charset=utf-8'):
        super().__init__(content)
        self._mime_tp = mime_type
        self._name = name

    def as_data_url(self, mime_type=None):
        """
            Returns the file contents as a data-uri (suitable to assign to
            a src attribute of images, href attribute of links, etc.)
        """
        if mime_type is None:
            mime_type = self._mime_tp
        return 'data:'+self._mime_tp+';base64,'+base64.b64encode(self.getvalue().encode('utf-8')).decode('ascii')

    def save(self):
        """
            Saves the file locally.

            It works by silently appending a link element to the body. The link element has an
            appropriate download attribute (the name of the file) and href attribute
            (the data-uri with the contents of the file). This element is then programmatically
            clicked and removed from the body.
        """
        link = html.A(href=self.as_data_url(), download=self._name)
        doc <= link
        link.click()
        doc.body.removeChild(link)


class FutureLocalFile(Future):
    def __init__(self, file):
        super().__init__()
        self._reader = window.FileReader.new()
        self._reader.bind('load', self._load_handler)
        self._reader.bind('error', self._error_handler)
        self._reader.readAsText(file)
        self._name = file.name
        self._mime = file.type

    def cancel(self):
        self._reader.abort()

    def _load_handler(self, evt):
        self.set_result(BrowserFile(self._name, evt.target.result, mime_type=self._mime))

    def _error_handler(self, evt):
        self.set_exception(IOError(str(self._reader.error)))


def open_local():
    """
    Opens a file dialog which the user can use to pick a file.

    Returns:
        asyncio.Future: Returns a future which, upon reading the file,
                        resolves to an `asyncio.fs.BrowserFile` object with
                        the file contents.

    The method works by appending an input element to the document body, registering a change
    handler for it, programmatically clicking it, and then immediately
    deleting it from the document.
    """

    i = html.INPUT(type='file')
    result = Future()

    @run_async()
    def change_handler(evt):
        ret = yield open(i.files[0])
        result.set_result(ret)

    i.bind('change', change_handler)
    doc <= i
    i.click()
    doc.body.removeChild(i)
    return result


@run_async()
def open(file, mode='r', size_limit=None):
    """
        Opens a local or remote file and returns a Future, which, upon completion,
        will resolve to an `asyncio.fs.BrowserFile` object with the file contents.

        Args:
          file(instance of the javascript File object/str):
                if opening a local file, the parameter must be an instance of the javascript File object
                if opening a remote file, the parameter should be the url of the file
          mode(str): unused, present for compatibility with the builtin open method
    """
    if 'r' in mode:
        if type(file) == str:
            req = yield HTTPRequest(file)
            if size_limit is not None and req.response.length > size_limit:
                raise IOError("Maximum file size exceeded:"+str(req.response.length)+"(max allowed:"+str(size_limit)+")")
            ret = BrowserFile(file, req.response)
        else:
            if size_limit is not None and file.size > size_limit:
                raise IOError("Maximum file size exceeded:"+str(file.size)+"(max allowed:"+str(size_limit)+")")
            ret = yield FutureLocalFile(file)
    else:
        ret = BrowserFile(file, '')
    return ret



from browser import ajax

from .futures import Future


class HTTPException(Exception):
    """
        A class representing a HTTPRequest error
    """
    def __init__(self, request):
        super(HTTPException, self).__init__()
        self.req = request


class HTTPRequest(Future):
    """
        A class representing a Future HTTPRequest result.
    """
    METHOD_POST = 'POST'
    METHOD_GET = 'GET'

    def __init__(self, url, method='GET', data=None, **kwargs):
        super(HTTPRequest, self).__init__(**kwargs)
        self._url = url
        self._req = ajax.ajax()
        self._req.bind("complete", self._complete_handler)
        self._data = data
        self._method = method
        self._req.open(self._method, self._url, True)
        self._req.set_header('content-type', 'application/x-www-form-urlencoded')
        if self._data is None:
            self._req.send()
        else:
            self._req.send(self._data)

    def cancel(self):
        self._req.abort()
        super().cancel()

    def _complete_handler(self, req):
        if req.status == 200 or req.status == 0:
            self.set_result(req)
        else:
            self.set_exception(HTTPException(req))



"""
Brython MDCFramework: Components
================================



"""


from .MDCComponent import MDCComponent

from .MDCButton import *
from .MDCDrawer import *
from .MDCCard import *
from .MDCChips import *
from .MDCDialog import *
from .MDCGridList import *
from .MDCLayoutGrid import *
from .MDCLinearProgress import *
from .MDCList import *
from .MDCMenu import *
from .MDCShape import *
from .MDCSnackbar import *
from .MDCTab import *
from .MDCTopAppBar import *
from .MDCFormField import *
from .MDCImageList import *



import os

files = os.listdir()
otfs = filter(lambda s: s.endswith('.otf'), files)

font_face = """
@font-face {{
  font-family: '{family}';
  font-style: {style};
  font-weight: {weight};
  src: url('/static/fonts/Ancizar-4.0/{name}') format('truetype');
}}
"""

names = []
with open('ancizar.css', 'w') as file:
    for i, name in enumerate(otfs):

        weight = name.split()[1].split("-")[1].split('.')[0].lower()
        family = name.split("-")[0].replace(' ', '-')
        style = 'italic' if 'Italic' in name else 'normal'
        weight = {
            'light': '400',
            'regular': '500',
            'bold': '700',
            'extrabold': '800',
            'black': '900',
        }[weight]

        file.write(font_face.format(**locals()))
        print(f"{i+1}: {name}")
        names.append(f'{family}-{style}-{weight}')


print('-' * 70)
print(', '.join(names) + ';')



from openbci.preprocess import eeg_filters
from openbci.preprocess import eeg_features
import logging
import inspect

import numpy as np


########################################################################
class EEGFrame:
    """"""

    # ----------------------------------------------------------------------
    def __init__(self, openbci):
        """Constructor"""

        if not(openbci.eeg_buffer.qsize() or openbci.eeg_pack.qsize()):
            logging.warning("Not enough data to creating `EEGFrame` object, EEG data must collected before.")
            return

        if openbci.eeg_buffer.qsize():
            self.data = np.array(openbci.eeg_buffer.queue)
            self.sample, self.eeg, self.aux, self.footer, self.timestamp = self.data.T

        elif openbci.eeg_pack.qsize():
            self.data = np.array(openbci.eeg_pack.queue)

        self.default_kwargs = {
            'sample': self.sample,
            # 'eeg': self.eeg,
            'aux': self.aux,
            'footer': self.footer,
            'timestamp': self.timestamp,
        }

    # ----------------------------------------------------------------------
    def __getattr__(self, attr):
        """"""
        if hasattr(eeg_filters, attr):
            _kwargs = {key: self.default_kwargs[key] for key in self.default_kwargs if key in inspect.getargspec(getattr(eeg_features, attr)).args}
            return lambda **kwargs: getattr(eeg_filters, attr)(np.stack(self.eeg), **{**_kwargs, **kwargs})

        elif hasattr(eeg_features, attr):
            _kwargs = {key: self.default_kwargs[key] for key in self.default_kwargs if key in inspect.getargspec(getattr(eeg_features, attr)).args}
            return lambda **kwargs: getattr(eeg_features, attr)(np.stack(self.eeg), **{**_kwargs, **kwargs})








"""
Brython MDCComponent: Core
==========================


"""

from browser import window, html

########################################################################


class MDCObject():
    """"""

    # ----------------------------------------------------------------------
    def __init__(self, view, mdc, element, name='MDCObject'):
        """Constructor"""
        self.__mdc__ = mdc
        self.__view__ = view
        self.__element__ = element
        # self.name_ = name
        # self.__repr__ = lambda: repr_

    # ----------------------------------------------------------------------
    def __getattr__(self, attr):
        """"""
        if hasattr(self.__view__, attr):  # check first for existence in view model, the method in view model can use MDC call
            def inset(*args, **kwargs):
                return getattr(self.__view__, attr)(self.__element__, *args, **kwargs)  # must be a classmethod
            return inset

        elif hasattr(self.__mdc__, attr):  # if not define in the view models then look up in MDC (foundation is not defined here)
            return getattr(self.__mdc__, attr)

    # ----------------------------------------------------------------------
    def __getitem__(self, item):
        """"""
        return self.__view__[item]  # views can define and use a __getitem__ as a shortcut

    # ----------------------------------------------------------------------

    # def __repr__(self):
        # """"""
        # return self.name_

    # # ----------------------------------------------------------------------
    # def __str__(self):
        # """"""
        # return self.name_


########################################################################
class MDCCore:
    """"""

    # ----------------------------------------------------------------------
    @classmethod
    def elevation(cls, element, Z):
        """"""
        cls.add_class(element, ['mdc-elevation--z{}'.format(Z), 'mdc-elevation-transition'])
        return element
        #cls.element.class_name += ' mdc-elevation--z{} mdc-elevation-transition'.format(Z)

    # ----------------------------------------------------------------------

    @classmethod
    def add_class(cls, element, classes):
        """"""
        # if not isinstance(classes, (list, tuple, set)):
            # classes = [classes]

        # print(cls.element.class_name)

        # print(cls)

        for class_ in classes:
            if class_ in cls.CSS_classes:
                class_ = cls.CSS_classes[class_]

            element.class_name += ' {}'.format(class_)

    # ----------------------------------------------------------------------

    @classmethod
    def remove_class(cls, element, classes):
        """"""
        # if not isinstance(classes, (list, tuple, set)):
            # classes = [classes]

        for class_ in classes:
            if class_ in cls.CSS_classes:
                class_ = cls.CSS_classes[class_]

            element.class_name = element.class_name.replace(class_, '')

    # ----------------------------------------------------------------------

    @classmethod
    def toggle_class(cls, element, class_):
        """"""
        if class_ in cls.CSS_classes:
            class_ = cls.CSS_classes[class_]

        if class_ in cls.element.class_name:
            cls.remove_class(element, cls.mdc, [class_])
        else:
            cls.add_class(element, cls.mdc, [class_])

    # ----------------------------------------------------------------------

    @classmethod
    def ripple(cls, element, theme='primary', unbounded=False):
        """"""

        if theme is 'secondary':
            theme = 'accent'

        cls.add_class(element, ['mdc-ripple-surface'])
        if theme in ['primary', 'accent']:
            cls.add_class(element, ['mdc-ripple-surface--{}'.format(theme)])

        if unbounded:
            cls.element.setAttribute('data-mdc-ripple-is-unbounded', '')

        ripple = window.mdc.ripple.MDCRipple.attachTo(element)
        # ripple.activate()

        # ripple = MDCObject(cls, ripple, element)

        return element

    # ----------------------------------------------------------------------

    @classmethod
    def theme(cls, element, theme, on=None):
        """"""
        if theme in ['primary', 'secondary', 'background', 'on-primary', 'on-secondary', 'secondary-bg', 'primary-bg']:
            cls.add_class(element, ['mdc-theme--{}'.format(theme)])

        elif theme in ['text-primary', 'text-secondary', 'text-hint', 'text-disabled', 'text-icon']:
            if on in ['light', 'dark']:
                cls.add_class(element, ['mdc-theme--{}-on-{}'.format(theme, on)])

        return element

    # ----------------------------------------------------------------------

    @classmethod
    def typography(cls, element, typo='typography'):
        """"""

        if typo == 'typography':
            cls.add_class(element, ['mdc-typography'])

        elif typo in ['headline1', 'headline2', 'headline3', 'headline4', 'headline5', 'headline6']:
            cls.add_class(element, ['mdc-typography--{}'.format(typo)])

        elif typo in ['subtitle1', 'subtitle2']:
            cls.add_class(element, ['mdc-typography--{}'.format(typo)])

        elif typo in ['body1', 'body2']:
            cls.add_class(element, ['mdc-typography--{}'.format(typo)])

        elif typo in ['caption', 'button', 'overline']:
            cls.add_class(element, ['mdc-typography--{}'.format(typo)])

        return element


########################################################################
class MDCTemplate(MDCCore):
    """"""

    NAME = None, None
    CSS_classes = {}
    MDC_optionals = {}

    # ----------------------------------------------------------------------

    def __new__(self, **kwargs):
        """"""
        self.element = self.render({}, kwargs)
        return self.element

    # ----------------------------------------------------------------------
    @classmethod
    def __html__(cls, **context):
        """"""
        code = """
            <div></div>
        """
        return cls.render_html(code, context)

    # ----------------------------------------------------------------------

    @classmethod
    def render(cls, locals_, kwargs, attach_now=True):
        """"""
        context = locals_.copy()
        context.pop('self')
        context.pop('kwargs')
        context.update(**kwargs)

        cls.html_element = cls.__html__(**context)

        if attach_now:
            cls.attach()

        if 'Class' in kwargs:
            cls.html_element.class_name += ' {Class}'.format(**kwargs)

        if 'id' in kwargs:
            cls.html_element.setAttribute('id', kwargs['id'])

        if 'style' in kwargs:
            if kwargs['style']:
                cls.html_element.style = kwargs['style']

        if 'attr' in kwargs:
            for attr in kwargs['attr']:
                cls.html_element.setAttribute(attr, kwargs['attr'][attr])

        return cls.html_element

    # ----------------------------------------------------------------------

    @classmethod
    def attach(cls):
        """"""
        html_element = cls.html_element
        name, mdcname = cls.NAME

        if name is None:
            html_element.mdc = MDCObject(cls, None, html_element)
            cls.mdc = html_element.mdc
            return

        try:
            cls.mdc = getattr(getattr(window.mdc, name), mdcname).attachTo(html_element)  # attach to MDC controller
            html_element.mdc = MDCObject(cls, cls.mdc, html_element)  # linked the html element with Views and MDCObject controller

            # foundation can be called from html element
            if hasattr(cls.mdc, 'foundation_'):
                html_element.foundation = cls.mdc.foundation_

        except Exception as e:
            # Some MDC objects like Buttons have not controllers (There is not a javascript controller)
            html_element.mdc = MDCObject(cls, None, html_element)
            cls.mdc = html_element.mdc

    # ----------------------------------------------------------------------
    @classmethod
    def render_html(cls, code, context):
        """"""

        classes = []
        for key in cls.CSS_classes.keys():
            if context.get(key, False):
                classes.append(cls.CSS_classes[key])
        context['CSS_classes'] = ' '.join(classes)

        for key in cls.MDC_optionals.keys():
            if context.get(key, False):

                try:
                    context[key] = cls.MDC_optionals[key].format(**context)
                except:

                    context[key] = cls.MDC_optionals[key]

            else:
                context[key] = ''

        code = code.format(**context)

        return cls.render_str(code)

    # ----------------------------------------------------------------------
    @classmethod
    def render_str(cls, code):
        """"""
        return html.DIV(code).children[0]

    # ----------------------------------------------------------------------

    @classmethod
    def new_id(cls):
        """"""
        ascii_lowercase = 'abcdefghijklmnopqrstuvwxyz'
        import random
        return ''.join([random.choice(ascii_lowercase) for l in range(2**5)])

    # ----------------------------------------------------------------------
    @classmethod
    def get_id(cls, element=None):
        """"""
        return cls.ID

    # ----------------------------------------------------------------------
    @classmethod
    def format_icon(cls, icon):
        """"""
        if icon and (icon.startswith('fa-') or icon.startswith('fas-') or icon.startswith('fab-') or icon.startswith('far-')):
            fa_style = icon[:icon.find('-')]
            fa_icon = 'fa' + icon[icon.find('-'):]
            return {'fa_icon': fa_icon,
                    'fa_icon_': fa_icon,
                    'fa_style': fa_style,
                    }

        return {'icon': icon,
                'icon_': icon,
                }



"""
Brython MDCComponent: MDCButton
===============================
"""
from .core import MDCTemplate


########################################################################
class MDCButton(MDCTemplate):
    """"""

    NAME = 'button', 'MDCButton'

    CSS_classes = {

        'raised': 'mdc-button--raised',
        'unelevated': 'mdc-button--unelevated',
        'outlined': 'mdc-button--outlined',
        'dense': 'mdc-button--dense',

    }

    MDC_optionals = {

        'disabled': 'disabled',
        'reversed': 'style = "margin-left: 8px; margin-right: -4px;"',

        # Icons
        'icon': '<i class="material-icons mdc-button__icon" {reversed} aria-hidden="true">{icon}</i>',
        'fa_icon': '<i class="mdc-button__icon {fa_style} {fa_icon}"></i>',

        # Icon for icon only
        'icon_': '<button class="mdc-icon-button material-icons">{icon_}</button>',
        'fa_icon_': '<button class="mdc-icon-button {fa_style} {fa_icon}"></button>',

        # Icons for a
        'icon_a': '<a class="mdc-icon-button material-icons">{icon_}</a>',
        'fa_icon_a': '<a class="mdc-icon-button {fa_style} {fa_icon}"></a>',


    }

    # ----------------------------------------------------------------------
    def __new__(self, text=None, href=None, icon=False, **kwargs):
        """"""
        kwargs.update(self.format_icon(icon))
        del icon

        # print("AAA", kwargs)

        self.element = self.render(locals(), kwargs)
        return self.element

    # ----------------------------------------------------------------------
    @classmethod
    def __html__(cls, **context):
        """"""
        if context.get('href'):
            if context.get('text'):
                if not context.get('reversed'):
                    code = """
                        <a href="{href}" class="mdc-button {CSS_classes}" {disabled}>
                        {icon}
                        {fa_icon}
                        {text}
                        </a>
                    """
                else:
                    code = """
                        <a href="{href}" class="mdc-button {CSS_classes}" {disabled}>
                        {text}
                        {icon}
                        {fa_icon}
                        </a>
                    """
            else:
                code = """
                {icon_a}
                {fa_icon_a}
                """

        else:
            if context.get('text'):
                if not context.get('reversed'):
                    code = """
                        <button class="mdc-button {CSS_classes}" {disabled}>
                        {icon}
                        {fa_icon}
                        {text}
                        </button>
                    """
                else:
                    code = """
                        <button class="mdc-button {CSS_classes}" {disabled}>
                        {text}
                        {icon}
                        {fa_icon}
                        </button>
                    """
            else:
                code = """
                    {icon_}
                    {fa_icon_}
                    """

        return cls.render_html(code, context)


########################################################################
class MDCFab(MDCButton):
    """"""

    CSS_classes = {

        'mini': 'mdc-fab--mini',
        'exited': 'mdc-fab--exited',
        'extended': 'mdc-fab--extended',

    }

    MDC_optionals = {

        'label': '<span class="mdc-fab__label">{label}</span>',
        'disabled': 'disabled',


        'icon': '<i class="material-icons mdc-fab__icon" aria-hidden="true">{icon}</i>',
        'fa_icon': '<i class="mdc-fab__icon {fa_style} {fa_icon}"></i>',

    }

    # ----------------------------------------------------------------------
    def __new__(self, icon, label=False, **kwargs):
        """"""
        kwargs.update(self.format_icon(icon))
        del icon

        if label:
            extended = True
        else:
            extended = False

        self.element = self.render(locals(), kwargs)
        return self.element

    # ----------------------------------------------------------------------
    @classmethod
    def __html__(cls, **context):
        """"""
        if not context.get('reversed'):
            code = """
                <button class="mdc-fab {CSS_classes}" {disabled}>
                  {icon}
                  {fa_icon}
                  {label}
                </button>
            """
        else:
            code = """
                <button class="mdc-fab {CSS_classes}" {disabled}>
                  {label}
                  {icon}
                  {fa_icon}
                </button>
            """

        return cls.render_html(code, context)


########################################################################
class MDCIconToggle(MDCTemplate):
    """"""

    NAME = 'iconToggle', 'MDCIconToggle'

    # ----------------------------------------------------------------------
    def __new__(self, icon_on, icon_off, state='off', **kwargs):
        """"""
        if state == 'on':
            state = icon_on
        elif state == 'off':
            state = icon_off

        self.element = self.render(locals(), kwargs)
        return self.element

    # ----------------------------------------------------------------------
    @classmethod
    def __html__(cls, **context):
        """"""
        code = """
        <i class="mdc-icon-toggle material-icons" role="button" aria-pressed="false"
           aria-label="Add to favorites" tabindex="0"
           data-toggle-on='{{"content": "{icon_on}"}}'
           data-toggle-off='{{"content": "{icon_off}"}}'>
           {state}
        </i>
        """

        return cls.render_html(code, context)



"""
Brython MDCComponent: MDCTopAppBar
==================================


"""


from .core import MDCTemplate
#from .MDCButton import MDCButton,  MDCIconToggle


# <a href="#" class="material-icons mdc-top-app-bar__action-item mdc-ripple-upgraded--unbounded mdc-ripple-upgraded" aria-label="Download" alt="Download" style="--mdc-ripple-fg-size:28.8px; --mdc-ripple-fg-scale:1.66667; --mdc-ripple-left:10px; --mdc-ripple-top:10px;">file_download</a>
# <a href="#" class="material-icons mdc-top-app-bar__action-item mdc-ripple-upgraded--unbounded mdc-ripple-upgraded" aria-label="Download" alt="Download" style="--mdc-ripple-fg-size:28.8px; --mdc-ripple-fg-scale:2.70424; --mdc-ripple-left:10px; --mdc-ripple-top:10px; --mdc-ripple-fg-translate-start:6.58438px, 9.6px; --mdc-ripple-fg-translate-end:9.6px, 9.6px;">file_download</a>


# <a href="#" class="material-icons mdc-top-app-bar__action-item mdc-ripple-upgraded--unbounded mdc-ripple-upgraded" aria-label="Download" alt="Download" style="--mdc-ripple-fg-size:28.8px; --mdc-ripple-fg-scale:1.66667; --mdc-ripple-left:10px; --mdc-ripple-top:10px;">file_download</a>
# <a href="#" class="material-icons mdc-top-app-bar__action-item mdc-ripple-upgraded--unbounded mdc-ripple-upgraded" aria-label="Download" alt="Download" style="--mdc-ripple-fg-size:28.8px; --mdc-ripple-fg-scale:1.66667; --mdc-ripple-left:10px; --mdc-ripple-top:10px;">file_download</a>


########################################################################
class __topAppBarItem__(MDCTemplate):
    """"""

    MDC_optionals = {

        'icon': '<a href="{href}" class="material-icons mdc-top-app-bar__action-item">{icon}</a>',
        'fa_icon': '<a href="{href}" class="material-icons mdc-top-app-bar__action-item"><i href="{href}" style="margin-left: 0px;" class="{fa_style} {fa_icon}"></i></a>',


    }

    # ----------------------------------------------------------------------
    def __new__(self, icon, href="#", **kwargs):
        """"""
        kwargs.update(self.format_icon(icon))
        del icon

        # if icon and icon.startswith('fa'):
            # fa_style = icon[:icon.find('-')]
            # fa_icon = 'fa' + icon[icon.find('-'):]
            # del icon

        self.element = self.render(locals(), kwargs)
        return self.element

    # ----------------------------------------------------------------------

    @classmethod
    def __html__(cls, **context):
        """"""
        code = """
            {icon}
            {fa_icon}
        """
        return cls.render_html(code, context)


########################################################################
class MDCTopAppBar(MDCTemplate):
    """"""

    NAME = 'topAppBar', 'MDCTopAppBar'

    CSS_classes = {

        # '_16_9': 'mdc-card__media--16-9',
        # 'square': 'mdc-card__media--square',
    }

    MDC_optionals = {

        # 'outlined': 'mdc-card--outlined',
        # 'full_bleed': 'mdc-card__actions--full-bleed',
        # 'icon': '<i class="material-icons mdc-button__icon" aria-hidden="true">{icon}</i>',
        # 'disabled': 'disabled',

        'short': 'mdc-top-app-bar--short mdc-top-app-bar--short-has-action-item',
        'collapsed': 'mdc-top-app-bar--short-collapsed',
        'fixed': 'mdc-top-app-bar--fixed',
        'prominent': 'mdc-top-app-bar--prominent',
        'dense': 'mdc-top-app-bar--dense',

        # 'menu_icon': '<a href="#" class="material-icons mdc-top-app-bar__navigation-icon">{menu_icon}</a>',
        # 'menu_icon': '<a href="#" class="baseline-{menu_icon} mdc-top-app-bar__navigation-icon"></a>',

        # Icons
        'icon': '<a class="material-icons mdc-top-app-bar__navigation-icon">{icon}</a>',
        'fa_icon': '<a class="mdc-top-app-bar__navigation-icon {fa_style} {fa_icon}"></a>',

    }

    # ----------------------------------------------------------------------

    def __new__(self, title, icon='menu', items=[], short=False, collapsed=False, prominent=False, fixed=True, dense=False, theme='primary', **kwargs):
        """"""

        kwargs.update(self.format_icon(icon))
        del icon

        self.element = self.render(locals(), kwargs)
        self.element.autodrawer = True

        for item in items:
            self.add_item(item)

        # self.element.bind('MDCTopAppBar:nav', lambda ev:print('hdhd'))

        return self.element

    # ----------------------------------------------------------------------

    @classmethod
    def __html__(cls, **context):
        """"""

        code = """

            <header class="mdc-top-app-bar app-bar mdc-theme--{theme}-bg mdc-theme--on-{theme} {short} {dense} {collapsed} {fixed} {prominent}" style='z-index: 1;'>
              <div class="mdc-top-app-bar__row">
                <section class="mdc-top-app-bar__section mdc-top-app-bar__section--align-start">
                  {icon}
                  {fa_icon}
                  <span class="mdc-top-app-bar__title">{title}</span>
                </section>
                <section class="mdc-top-app-bar__section mdc-top-app-bar__section--align-end" role="toolbar">
                </section>
              </div>
            </header>
        """

        return cls.render_html(code, context)

    # ----------------------------------------------------------------------

    @classmethod
    def __getitem__(self, name):
        """"""
        if name is 'actions':
            return self.element.select('.mdc-card__actions')[0]

        elif name is 'section_start':
            return self.element.select('.mdc-top-app-bar__section--align-start')[0]

        elif name is 'icons_parent':
            return self.element.select('.mdc-top-app-bar__section--align-end')[0]

        elif name is 'icon':
            return self.element.select('.mdc-top-app-bar__navigation-icon')[0]

        elif name is 'title':
            return self.element.select('.mdc-top-app-bar__title')[0]

    # ----------------------------------------------------------------------

    @classmethod
    def add_item(cls, element, *args, **kwargs):
        """"""

        item = __topAppBarItem__(*args, **kwargs)
        # print(cls['icons_parent'])
        cls['icons_parent'] <= item

        return item

    # ----------------------------------------------------------------------

    @classmethod
    def title(self, element, text):
        """"""
        self['title'].html = text




from .futures import Future, CancelledError
from ._utils import decorator, _isgenerator

import types
import dis

@decorator
def coroutine(func):
    """
        Implementation adapted from https://www.pythonsheets.com/notes/python-asyncio.html
    """
    func.__coroutinefunction__ = True
    return func

"""
    def coro(*a, **k):
        res = func(*a, **k)
        if isinstance(res, Future) or _isgenerator(res):
            yield from res
        else:
            return res
    coro.__coroutinefunction__ = True
    return coro
"""

def run_async(loop=None):
    @decorator
    def _decorator(func):
        coro = coroutine(func)
        def task(*a, **k):
            return Task(coro(*a, **k), loop=loop, task_name=func.__name__)
        task.__async_task__ = True
        return task
    return _decorator


def iscoroutinefunction(fun):
    return (hasattr(fun, '__coroutinefunction__') or
        fun.__code__.co_flags & dis.COROUTINE)

def iscoroutine(obj):
    return isinstance(obj, types.CoroutineType) or _isgenerator(obj)


class Task(Future):
    """
        Implementation adapted from https://www.pythonsheets.com/notes/python-asyncio.html
    """
    def __init__(self, coro_object, *, loop=None, task_name=None):
        super().__init__(loop)
        self._coro_obj = coro_object
        self._name = task_name
        self._loop.call_soon(self._step)

    def _step(self, val=None, exc=None):
        if self.done():
            return
        try:
            if exc:
                f = self._coro_obj.throw(exc)
            else:
                f = self._coro_obj.send(val)
        except StopIteration as e:
            self.set_result(e.value)
        except BaseException as e:
            self.set_exception(e)
        else:
            f.add_done_callback(self._wakeup)

    def _wakeup(self, fut):
        try:
            res = fut.result()
        except BaseException as e:
            self._step(None, e)
        else:
            self._step(res, None)

    def __str__(self):
        ret = "Task("+str(self._name)+"): "
        if self.done():
            if self._exception:
                ret += "finished with exception"
            else:
                ret += "finished with result:"+str(self._result)
        elif self.cancelled():
            ret += "canceled"
        else:
            ret += "pending"
        return ret


def ensure_future(fut_or_coroutine_obj, *, loop=None):
    if isinstance(fut_or_coroutine_obj, Future):
        if loop is not None and loop is not fut_or_coroutine_obj._loop:
            raise ValueError('loop argument must agree with Future')
        return fut_or_coroutine_obj
    elif iscoroutine(fut_or_coroutine_obj):
        return Task(fut_or_coroutine_obj, loop=loop,
            task_name=fut_or_coroutine_obj.__name__)
    else:
        raise TypeError('Expecting coroutine object or Future, got ' +
            str(fut_or_coroutine_obj) + ' instead.')




import types

def decorator(dec):
    def new_dec(fn):
        ret = dec(fn)
        ret.__decorated = fn
        ret.__name__ = fn.__name__
        return ret
    return new_dec

def _isgenerator_function(val):
    return hasattr(val,'__code__') and val.__code__.co_flags & 0x20

def _isgenerator(val):
    return isinstance(val, types.GeneratorType)





"""
================
EEG Connectivity
================
"""

import numpy as np
from sklearn.linear_model import LinearRegression


# ----------------------------------------------------------------------
def gc_connectivity(data, order):
    """ Granger Causality
    
    The function gc_connectivity computes the Granger Causality of a given 
    order among the time series contained in the rows of the matrix data.
    
    Parameters
    ----------
    data : matrix
        Matrix whose rows contain the time series for which GC will be 
        computed (channels x N).
    order : int
        Order of the autoregressive models used to compute GC.
    
    Returns
    -------
    matrix
        Granger causality matrix (channels x channels)
    """

    # Number of channels or times series (rows in the matrix data)
    num_ch = np.shape(data)[0]
    GC_matrix = np.zeros((num_ch, num_ch))

    for i in range(0, num_ch):
        for j in range(0, num_ch):
            if i != j:
                x = data[i]
                y = data[j]
                GC_matrix[i, j] = granger_causality(x, y, order)
    return GC_matrix


# ----------------------------------------------------------------------
def granger_causality(x, y, order):
    """Granger Causality for x and y.
    
    The function granger_causality computes the Granger Causality of a given order
    between the time series x and y, assuming that x drives y.
    
    Parameters
    ----------
    x : int
         Driving time series (1 x N).
    y : str
        Driven time series (1 x N).
    order: int
        Order of the autoregressive models used to compute GC.
    
    Returns
    -------
    float
        Granger causality between x and y.
    """

    X = np.zeros((len(x) - order, order))
    Y = np.zeros((len(y) - order, order))

    for p in range(0, order):
        X[:, p] = x[p:-(order - p)]
        Y[:, p] = y[p:-(order - p)]

    A0 = np.ones((np.shape(Y)[0], 1))
    Y_aux = y[order::]
    Y_aux = np.resize(Y_aux, (len(Y_aux), 1))

    # Autoregressive model
    regressor_y = LinearRegression()
    regressor_y.fit(np.concatenate((A0, Y), axis=1), Y_aux)
    coeff_y = regressor_y.coef_  # Regression coefficients
    u_y = Y_aux - np.dot(coeff_y, np.concatenate((A0, Y), axis=1).T).T  # Residuals

    # Bivariate autoregressive model
    regressor_yx = LinearRegression()
    regressor_yx.fit(np.concatenate((A0, Y, X), axis=1), Y_aux)
    coeff_yx = regressor_yx.coef_  # Regression coefficients
    u_yx = Y_aux - np.dot(coeff_yx, np.concatenate((A0, Y, X), axis=1).T).T  # Residuals

    GC = np.log(np.var(u_y) / np.var(u_yx))  # Granger Causality

    if GC < 0:
        GC = 0.0

    return GC



"""
Brython MDCComponent: MDCImageList
==================================


"""


from .core import MDCTemplate
#from .MDCButton import MDCButton,  MDCIconToggle


########################################################################
class __listItem__(MDCTemplate):
    """"""

    MDC_optionals = {

        'columns': 'width: calc(100% / {columns} - 2*{margin}px);',
        'margin': 'margin: {margin}px;',
        'margin_bottom': 'margin-bottom: {margin_bottom};',
        'label': '<div class="mdc-image-list__supporting"><span class="mdc-image-list__label">{label}</span></div>',


    }

    # ----------------------------------------------------------------------

    def __new__(self, columns, margin, masonry, src, label='', **kwargs):
        """"""
        if masonry:
            margin_bottom = margin
            columns = False
            margin = False

        self.element = self.render(locals(), kwargs)
        return self.element

    # ----------------------------------------------------------------------

    @classmethod
    def __html__(cls, **context):
        """"""

        if context['masonry']:
            code = """

              <li class="mdc-image-list__item image_container" style='{margin_bottom}'>
                <img class="mdc-image-list__image" src="{src}">
                {label}
              </li>

            """
        else:
            code = """

              <li class="mdc-image-list__item" style='{columns} {margin}'>
                <div class="mdc-image-list__image-aspect-container image_container" style="padding-bottom: 66.66667%;">
                  <img class="mdc-image-list__image" src="{src}" style="width: unset !important;">
                </div>
                {label}
              </li>

            """

        return cls.render_html(code, context)

    # ----------------------------------------------------------------------

    @classmethod
    def __getitem__(self, name):
        """"""
        if name is 'image_container':
            return self.element.select('.image_container')[0]

        elif name is 'label':
            return self.element.select('.mdc-image-list__supporting')[0]

        elif name is 'image':
            return self.element.select('.mdc-image-list__image')[0]

        # elif name is 'action_icons':
            # return self.element.select('.mdc-card__action-icons')[0]


########################################################################
class MDCImageList(MDCTemplate):
    """"""

    #NAME = 'LinearProgress'

    MDC_optionals = {

        # '_16_9': 'mdc-card__media--16-9',
        # 'square': 'mdc-card__media--square',
        'masonry': 'mdc-image-list--masonry masonry-image-list',
        'text_protection': 'mdc-image-list--with-text-protection',

        'masonry_column': 'column-count: {masonry_column};',
        'masonry_margin': 'column-gap: {masonry_margin};',


    }

    # ----------------------------------------------------------------------

    def __new__(self, columns=5, margin='2px', masonry=False, text_protection=False, **kwargs):
        """"""

        self.columns = columns
        self.margin = margin
        self.masonry = masonry

        if masonry:
            masonry_column = columns
            masonry_margin = margin
        else:
            masonry_column = False
            masonry_margin = False

        self.element = self.render(locals(), kwargs)
        return self.element

    # ----------------------------------------------------------------------

    @classmethod
    def __html__(cls, **context):
        """"""

        code = """
        <ul class="mdc-image-list {masonry} {text_protection}" style="{masonry_column} {masonry_margin}">
        </ul>
        """

        return cls.render_html(code, context)

    # ----------------------------------------------------------------------

    @classmethod
    def __getitem__(self, name):
        """"""
        # if name is 'actions':
            # return self.element.select('.mdc-card__actions')[0]

        # elif name is 'action_buttons':
            # return self.element.select('.mdc-card__action-buttons')[0]

        # elif name is 'action_icons':
            # return self.element.select('.mdc-card__action-icons')[0]

    # ----------------------------------------------------------------------
    @classmethod
    def add_item(cls, element, **kwargs):
        """"""
        item = __listItem__(cls.columns, cls.margin, cls.masonry, **kwargs)
        cls.element <= item
        return item



"""
Brython MDCComponent: MDCSnackbar
=================================


"""

from .core import MDCTemplate
#from .MDCButton import MDCButton,  MDCIconToggle


########################################################################
class MDCSnackbar(MDCTemplate):
    """"""

    NAME = 'snackbar', 'MDCSnackbar'

    CSS_classes = {

        #'_16_9': 'mdc-card__media--16-9',
        #'square': 'mdc-card__media--square',
    }

    MDC_optionals = {

        #'outlined': 'mdc-card--outlined',
        #'full_bleed': 'mdc-card__actions--full-bleed',
        #'icon': '<i class="material-icons mdc-button__icon" aria-hidden="true">{icon}</i>',
        #'disabled': 'disabled',

        'align_start': 'mdc-snackbar--align-start',

    }

    #----------------------------------------------------------------------
    def __new__(self, align_start=False, **kwargs):
        """"""
        self.element = self.render(locals(), kwargs)
        return self.element



    #----------------------------------------------------------------------
    @classmethod
    def __html__(cls, **context):
        """"""
        code = """
            <div class="mdc-snackbar {align_start}" style="z-index: 999999"
                 aria-live="assertive"
                 aria-atomic="true"
                 aria-hidden="true">
              <div class="mdc-snackbar__text"></div>
              <div class="mdc-snackbar__action-wrapper">
                <button type="button" class="mdc-snackbar__action-button"></button>
              </div>
            </div>
        """
        return cls.render_html(code, context)




    #----------------------------------------------------------------------
    @classmethod
    def show(cls, element, message, action_text=None, action_handler=None, timeout=2750, multiline=False, actiononbottom=False):
        """"""

        if action_text and not action_handler:
            action_handler = lambda :None

        obj = {
            'message': message,
            'timeout': timeout,
            'actionHandler': action_handler,
            'actionText': action_text,
            'multiline': multiline,
            'actionOnBottom': actiononbottom,
        }

        cls.mdc.show(obj)


    #----------------------------------------------------------------------
    @classmethod
    def __getitem__(self, name):
        """"""
        if name is 'text':
            return self.element.select('.mdc-snackbar__text')[0]

        #elif name is 'action_buttons':
            #return self.element.select('.mdc-card__action-buttons')[0]

        #elif name is 'action_icons':
            #return self.element.select('.mdc-card__action-icons')[0]


    #----------------------------------------------------------------------
    @classmethod
    def add_action_button(cls, element, mdc, *args, **kwargs):
        """"""



    ##----------------------------------------------------------------------
    #@classmethod
    #def title(self, mdc, text):
        #""""""
        #self['title'].text = text


